/* 
 * RC-like remote control over UDP/IPv4; transmitter part.
 * Copyright (c) 2005, 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

/* linux/input.h (included from joystick.h) is broken in some kernels */
#ifndef BITS_PER_LONG
#define BITS_PER_LONG 32
typedef unsigned long kernel_ulong_t;
#endif

#include <linux/joystick.h>

void fatal(char *msg) {
  if ( errno ) perror(msg);
  else fprintf(stderr, "%s\n", msg);
  exit(1);
}

#define NAME_LENGTH 128

#define N_CHANS 6
#define N_AXES 6

#define BUTTON_TRIM 9
#define BUTTON_CLEARTRIM 4
#define BUTTON_AUTOPILOT 0

void save_trims(char *trim_file, int *trims) {
  int a;
  FILE *f = fopen(trim_file, "wc");
  if ( f ) {
    for ( a=0; a<N_AXES; ++a ) fprintf(f, "%d ", trims[a]);
    fprintf(f, "\n");
    fclose(f);
    printf("Trims recorded.\n"); fflush(stdout);
  }
}

int main (int argc, char **argv) {
  char *mix_file = "xpad_linear6.mix";
  char *trim_file = ".trim";
  char *dev = "/dev/js0";
  char *udp_addr = "localhost";
  unsigned short udp_port = 9001;
  int fd;
  int sock;
  struct sockaddr_in dst;
  int rate = 50;

  int mix_matrix[N_CHANS][N_AXES];
  int mix_base[N_CHANS], mix_scale[N_CHANS];
  char mix_type[N_CHANS];
  char mix_trimmable[N_AXES];
  int mix_trim[N_AXES];

  unsigned char n_axes;
  unsigned char n_buttons;

  {
    int i;
    
    for ( i=1; i<argc; ++i ) {
      if      ( !strcmp(argv[i], "-c") && ++i<argc ) dev = argv[i];
      else if ( !strcmp(argv[i], "-d") && ++i<argc ) udp_addr = argv[i];
      else if ( !strcmp(argv[i], "-p") && ++i<argc ) udp_port = atoi(argv[i]);
      else if ( !strcmp(argv[i], "-m") && ++i<argc ) mix_file = argv[i];
      else if ( !strcmp(argv[i], "-r") && ++i<argc ) rate = atoi(argv[i]);
      else fatal("Usage:");
    }
  }
  
  {
    FILE *f = fopen(mix_file, "r");
    int c, a;
    char dummy[256];
    if ( !f ) fatal(mix_file);
    for ( c=0; c<N_CHANS; ++c ) {
      float v, w;
      if ( fscanf(f, "%c %f %f", &mix_type[c], &v, &w) != 3 )
	fatal("Invalid mixer file format");
      mix_base[c] = v * 65536;
      mix_scale[c] = w * 256;
      for ( a=0; a<N_AXES; ++a ) {
	if ( fscanf(f, "%f", &v) != 1 ) fatal("Invalid mixer file format");
	mix_matrix[c][a] = v * 256;
      }
      fgets(dummy, sizeof(dummy)-1, f);
    }
    for ( a=0; a<N_AXES; ++a ) {
      char c;
      if ( fscanf(f, " %c", &c) != 1 || (c!='Y' && c!='N') )
	fatal("Invalid mixer file format");
      mix_trimmable[a] = (c=='Y');
      mix_trim[a] = 0;
    }
    fclose(f);
    printf("Using mixer configuration %s\n", mix_file);
  }

  {
    int a;
    FILE *f = fopen(trim_file, "r");
    if ( f ) {
      for ( a=0; a<N_AXES; ++a ) {
	int offs;
	if ( fscanf(f, "%d", &offs) != 1 ) fatal("Invalid trim file format");
	mix_trim[a] = offs;
      }
      fclose(f);
    }
  }
	  

  {
    int vers;
    char name[NAME_LENGTH] = "";

    fd = open(dev, O_RDONLY);
    if ( fd < 0 ) fatal(dev);
    
    if ( ioctl(fd, JSIOCGVERSION, &vers) < 0 ) fatal("JSIOCGVERSION");
    if ( ioctl(fd, JSIOCGAXES,    &n_axes)    < 0 ) fatal("JSIOCGAXES");
    if ( ioctl(fd, JSIOCGBUTTONS, &n_buttons) < 0 ) fatal("JSIOCGBUTTONS");
    if ( ioctl(fd, JSIOCGNAME(NAME_LENGTH), name) < 0 ) fatal("JSIOCGNAME");
    
    printf("Joystick %s (%s), %d axes, %d buttons, version %d.%d.%d.\n",
	   dev, name, n_axes, n_buttons,
	   vers>>16, (vers>>8)&0xff, vers&0xff);
  }

  sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if ( sock < 0 ) fatal("socket");

  {
    struct hostent *he = gethostbyname(udp_addr);
    if ( ! he ) fatal(udp_addr);
    memcpy(&dst.sin_addr.s_addr, he->h_addr_list[0], he->h_length);
    dst.sin_family = AF_INET;
    dst.sin_port = htons(udp_port);
    printf("Sending to %s:%d.\n", udp_addr, udp_port);
    printf("Refresh rate %d Hz.\n", rate);
  }

  {
    struct js_event js; 
    struct timeval now, latest;
    fd_set fds;
    int *axes = malloc(n_axes * sizeof(int));
    int *buttons = malloc(n_buttons * sizeof(int));
    
    if ( !axes || !buttons ) fatal("malloc");

    if ( gettimeofday(&latest, NULL) < 0 ) fatal("gettimeofday");

    FD_ZERO(&fds);

    while ( 1 ) {
      struct timeval tv;
      tv.tv_sec = 0;
      tv.tv_usec = 1000000 / rate;
      FD_SET(fd, &fds);
      if ( select(fd+1, &fds, NULL, NULL, &tv) < 0 ) fatal("select");
      
      if ( FD_ISSET(fd, &fds) ) {
	if ( read(fd, &js, sizeof(js)) != sizeof(js) ) fatal("read");
	switch ( js.type & ~JS_EVENT_INIT ) {
	case JS_EVENT_BUTTON:
	  if ( js.number==BUTTON_TRIM && !buttons[js.number] && js.value ) {
	    int a;
	    for ( a=0; a<N_AXES; ++a ) {
	      mix_trim[a] += axes[a];
	    }
	    save_trims(trim_file, mix_trim);
	  }
	  else if ( js.number==BUTTON_CLEARTRIM && js.value ) {
	    int a;
	    for ( a=0; a<N_AXES; ++a ) mix_trim[a] = 0;
	    save_trims(trim_file, mix_trim);
	  }	    
	  buttons[js.number] = js.value;
	  break;
	case JS_EVENT_AXIS:
	  axes[js.number] = js.value;
	  break;
	}
      }

      if ( gettimeofday(&now, NULL) < 0 ) fatal("gettimeofday");
      int dt = (now.tv_sec-latest.tv_sec)*1000000
	+ (now.tv_usec-latest.tv_usec);
      if ( dt*rate >= 1000000 ) {
	int c, a;
	struct {
	  uint32_t flags;
	  uint16_t chans[N_CHANS];
	} msg;
	for ( c=0; c<N_CHANS; ++c ) {
	  int v = 0;
	  for ( a=0; a<N_AXES; ++a ) {
	    int val = axes[a];
	    if ( mix_trimmable[a] ) {
	      if ( buttons[BUTTON_TRIM] ) val = 0;
	      val += mix_trim[a];
	    }
	    v += (mix_matrix[c][a] * val) >> 8;
	  }
	  v = mix_base[c] + ((v*mix_scale[c]) >> 8);
	  if ( v < 0 ) v = 0;
	  else if ( v > 65535 ) v = 65535;
	  msg.chans[c] = htons(v);
	}
	msg.flags = buttons[BUTTON_AUTOPILOT] ? 1 : 0;
	msg.flags = htonl(msg.flags);
	if ( sendto(sock, &msg, sizeof(msg), 0,
		    (struct sockaddr*)&dst, sizeof(dst)) < 0 )
	  fatal("sendto");
	latest = now;
      }
    }
  }

  return 0;
}
