/* -*- linux-c -*-
 *
 * Cache management functions, as per the PXA255 user's manual.
 * Copyright (c) 2005, 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/dma-mapping.h>

static void do_lock_i_cache(void *begin, void *end) {
	register u32 tmp;
	unsigned long flags;
	local_irq_save(flags);
	__asm__ volatile (
		".macro cpwait1                   \n\
                mrc p15, 0, %[tmp], c2, c0, 0     \n\
                mov %[tmp], %[tmp]                \n\
                sub pc, pc, #4                    \n\
                .endm                             \n\
                @ Invalidate icache               \n\
                mcr p15, 0, %[tmp], c7, c5, 0     \n\
                cpwait1                           \n\
		loop_iline:                       \n\
                mcr p15, 0, %[addr], c9, c1, 0    \n\
                add %[addr], %[addr], #32         \n\
                cmp %[addr], %[end]               \n\
                bne loop_iline"
		: [addr]"=&r"(begin), [end]"=&r"(end), [tmp]"=&r"(tmp)
		: "0"((u32)begin&~31), "1"((((u32)end-1)|31)+1));
	local_irq_restore(flags);
}
static void do_lock_i_cache_end(void) { }

static void lock_i_cache(void *begin, void *end) {
	dma_addr_t handle;
	void (*p)(void *begin, void *end);
	int len = (void*)do_lock_i_cache_end - (void*)do_lock_i_cache;
	if ( end <= begin ) { printk("invalid lock range\n"); return; }
	printk("Locking %d bytes into i-cache\n", end-begin);
	p = dma_alloc_coherent(NULL, 4096, &handle, GFP_KERNEL|GFP_DMA);
	if ( !p ) { printk("lock_i_cache failed\n"); return; }
	memcpy(p, do_lock_i_cache, len); 
	flush_icache_range((long)p, (long)p+len);
	(*p)(begin, end);
	dma_free_coherent(NULL, 4096, p, handle);
}

static void unlock_i_cache(void) {
  __asm__ volatile ("mcr p15, 0, r0, c9, c1, 1");
}

/* */

static void do_lock_i_tlb(void *addr) {
	register u32 tmp;
	unsigned long flags;
	local_irq_save(flags);
	__asm__ volatile (
		".macro cpwait2                   \n\
                mrc p15, 0, %[tmp], c2, c0, 0     \n\
                mov %[tmp], %[tmp]                \n\
                sub pc, pc, #4                    \n\
                .endm                             \n\
                @ Invalidate then lock.           \n\
                mcr p15, 0, %[addr], c8, c5, 1    \n\
                mcr p15, 0, %[addr], c10, c4, 0   \n\
                cpwait2"
		: [tmp]"=&r"(tmp)
		: [addr]"r"(addr));
	local_irq_restore(flags);
}
static void do_lock_i_tlb_end(void) { }

static void lock_i_tlb(void *begin, void *end) {
	dma_addr_t handle;
	void (*p)(void *addr);
	int len = (void*)do_lock_i_tlb_end - (void*)do_lock_i_tlb;
	p = dma_alloc_coherent(NULL, 4096, &handle, GFP_KERNEL|GFP_DMA);
	if ( !p ) { printk("lock_i_tlb failed\n"); return; }
	memcpy(p, do_lock_i_tlb, len); 
	begin = (void*)((u32)begin & PAGE_MASK); // TBD 1k ?
	while ( begin < end ) {
		printk("Locking %p into i-tlb\n", begin);
		(*p)(begin);
		begin += PAGE_SIZE;
	}
	dma_free_coherent(NULL, 4096, p, handle);
}

static void unlock_i_tlb(void) {
	__asm__ volatile ("mcr p15, 0, r0, c10, c4, 1");
}

/* */

static void lock_d_cache(void *begin, void *end) {
	register u32 tmp;
	if ( end <= begin ) {
		printk(KERN_WARNING "begin=%p end=%p\n", begin, end);
		return;
	}
	printk("Locking %d bytes into d-cache\n", end-begin);
	unsigned long flags;
	local_irq_save(flags);
	__asm__ volatile (
		".macro drain3                    \n\
                mcr p15, 0, %[tmp], c7, c10, 4    \n\
                .endm                             \n\
                .macro cpwait3                    \n\
                mrc p15, 0, %[tmp], c2, c0, 0     \n\
                mov %[tmp], %[tmp]                \n\
                sub pc, pc, #4                    \n\
                .endm                             \n\
                drain3                            \n\
                @ Enter lock mode                 \n\
                mov %[tmp], #1                    \n\
                mcr p15, 0, %[tmp], c9, c2, 0     \n\
                cpwait3                           \n\
                loop_dline:                       \n\
                mcr p15, 0, %[addr], c7, c10, 1   \n\
                mcr p15, 0, %[addr], c7, c6, 1    \n\
                @ pld doesn't work as advertised  \n\
                @ TBD Needs an erratum ?          \n\
                @pld [%[addr]]                    \n\
                ldr %[tmp], [%[addr]]             \n\
                add %[addr], %[addr], #32         \n\
                drain3                            \n\
                cmp %[addr], %[end]               \n\
                bne loop_dline                    \n\
                drain3                            \n\
                mov %[tmp], #0                    \n\
                @ Leave lock mode                 \n\
                mcr p15, 0, %[tmp], c9, c2, 0     \n\
                cpwait3"
		: [addr]"=&r"(begin), [tmp]"=&r"(tmp)
		: "0"((u32)begin&~31), [end]"r"((((u32)end-1)|31)+1));
	local_irq_restore(flags);
}

static void unlock_d_cache(void) {
  __asm__ volatile ("mcr p15, 0, r0, c9, c2, 1");
}

/* */

static void lock_d_tlb(void *begin, void *end) {
	u32 tmp;
	unsigned long flags;
	local_irq_save(flags);
	begin = (void*)((u32)begin & PAGE_MASK); // TBD 1k ?
	while ( begin < end ) {
		printk("Locking %p into d-tlb\n", begin);
		__asm__ volatile (
		  ".macro cpwait4                   \n\
                  mrc p15, 0, %[tmp], c2, c0, 0     \n\
                  mov %[tmp], %[tmp]                \n\
                  sub pc, pc, #4                    \n\
                  .endm                             \n\
                  @ Invalidate then lock.           \n\
                  mcr p15, 0, %[addr], c8, c6, 1    \n\
                  cpwait4  @ TBD Needs an erratum ? \n\
                  mcr p15, 0, %[addr], c10, c8, 0   \n\
                  cpwait4"
		  : [tmp]"=&r"(tmp)
		  : [addr]"r"(begin));
		begin += PAGE_SIZE;
	}
	local_irq_restore(flags);
}

static void unlock_d_tlb(void) {
	__asm__ volatile ("mcr p15, 0, r0, c10, c8, 1");
}

