/* -*- linux-c -*-
 *
 * Driver for a CMOS sensor connected to the PCMCIA/CF bus of the
 * PXA255 and clocked by nPIOR.
 * Copyright (c) 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/vmalloc.h>

#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/leds.h>
#include <asm/irq.h>
#include <asm/fiq.h>
#include <asm/mach/irq.h>
#include <asm/mach/time.h>
#include <asm/arch/pxa-regs.h>
#include <asm/cacheflush.h>

#include <asm/arch/dma.h>

#include <linux/dma-mapping.h>

#include "pxa_cfcam.h"

#define NAME "cfcam"

#define NOIRQ 0

/* Module API */

MODULE_DESCRIPTION("Driver for CMOS sensor on PXA255 PCMCIA/CF bus");
MODULE_AUTHOR("pascal@pabr.org");
MODULE_LICENSE("GPL");

static int dev_major = 0;
module_param(dev_major, int, 0400);
MODULE_PARM_DESC(dev_major, "Device major number (default: auto-alloc)");

static unsigned long addr = 0x20000000;
module_param(addr, ulong, 0400);
MODULE_PARM_DESC(addr, "Physical address to read from");

static int sensor = 0;
module_param(sensor, int, 0400);
MODULE_PARM_DESC(sensor, "Sensor number");

static int gpio_reset = GPIO8_MMCCS0;
module_param(gpio_reset, int, 0400);
MODULE_PARM_DESC(gpio_reset, "GPIO pin for sensor reset");
static int gpio_powerdown = GPIO55_nPREG;
module_param(gpio_powerdown, int, 0400);
MODULE_PARM_DESC(gpio_powerdown, "GPIO pin for sensor power down");

static int gpio_sda = GPIO52_nPCE_1;
module_param(gpio_sda, int, 0400);
MODULE_PARM_DESC(gpio_sda, "GPIO pin for sensor I2C SDA");
static int gpio_scl = GPIO53_nPCE_2;
module_param(gpio_scl, int, 0400);
MODULE_PARM_DESC(gpio_scl, "GPIO pin for sensor I2C SCL");

struct cfcam_sensor {
	const char *name;
	int w;           // width of useful area
	int h;           // height of useful area
	int bits;        // significant bits per pixel
	int bpp;         // bytes read per pixel
	int preskip;     // clocks from reset to start of frame
	int skip;        // bytes to skip at beginning of frame
	int vskip;       // lines to skip at beginning of frame
	int hskip;       // bytes to skip at beginning of line
	char *i2c_cfg;   // NULL if no I2C
	int i2c_addr;    // ignored if no I2C
	// Internal use
	int i2c_skipped; // clocks done during i2c init
	volatile char *mem; // uncached map for clocking
};

static struct cfcam_sensor sensors[] = {
	{ .name = "Raw data (1 pixel per clock, no hsync, no vsync)",
	  .w = 1700,
	  .h = 600,
	  .bits = 9,
	  .bpp = 2,
	  .preskip = 0, // default yflip, default IT
	  .skip = 0,
	  .vskip = 0,
	  .hskip = 0,
	},
	{ .name = "HV7131R default (VGA)",
	  .w = 642,
	  .h = 482,
	  .bits = 9,
	  .bpp = 4,
	  .preskip = (850*477+820)*2, // default yflip, default IT
	  .skip = 668,
	  .vskip = 16,
	  .hskip = 208 * 4,
	},
	{ .name = "HV7131R with I2C config (VGA, yflip) - BROKEN",
	  .w = 642,
	  .h = 482,
	  .bits = 9,
	  .bpp = 4,
	  .preskip = (850*18+772)*2, // -yflip, IT=0x008000
	  .skip = 668,
	  .vskip = 16,
	  .hskip = 208 * 4,
	  .i2c_addr = 0x22,
	  .i2c_cfg = "0x01:0x08"
	},
	{ .name = "HV7131R with I2C config (CIF, yflip)",
	  .w = 322,
	  .h = 242,
	  .bits = 9,
	  .bpp = 4,
	  .preskip = 79772 + 270720*3, // - 2*((322+208)*10+10),
	  .skip = 840,
	  .vskip = 13,
	  .hskip = 208 * 4,
	  .i2c_addr = 0x22,
	  // Note: HBLANK<208 is ignored by the sensor
	  .i2c_cfg =
	  "0x01:0x08,"                     // YFlip
	  "0x11:122,0x13:162,"             // Start (162,122)
	  //"0x21:208,0x23:8,"             // HBLANK, VBLANK
	  "0x14:0,0x15:242,"               // Height 240+2
	  "0x16:1,0x17:66,"                // Width 320+2
	  "0x25:0x00,0x26:0x80,0x27:0x00," // Exposure
	  "0x30:0xf0"                      // Gain
	},
};

struct cfcam_dev {
	struct semaphore opened;
	// Config
	int dma_chan;
	struct cfcam_sensor *sens;
	// Buffers
	//// DMA buf
	int dma_size;
	dma_addr_t h_dma_buf;
	char *dma_buf;
	char **lines;  // Points into dma_buf
	//// DMA descriptors
	int descs_size;
	dma_addr_t h_descs;
	struct pxa_dma_desc *descs;
	// State
	struct pxa_cfcam_transform *tfm; // After open()
	int current_frame;
	enum { IDLE, SYNCING, READING, COMPLETE, MMAP } state;
	int prev_line;     // For SYNCING
	char __user *dst;         // For SYNCING, READING, MMAP
	int remain;        // For SYNCING, READING, MMAP
	//wait_queue_head_t wait; // For SYNCING, READING
	// MMAP API
	char *kbuf;     // After open(), NULL except for MMAP
	int kbuf_size;  // For MMAP
	int kbuf_cap;   // For MMAP
	struct vm_area_struct *kbuf_vma;
};

// Returns the line currently being written by DMA.

int current_line(struct cfcam_dev *cam) {
	u32 ddadr = DDADR(cam->dma_chan);
	int desc = (ddadr-cam->h_descs) / sizeof(struct pxa_dma_desc);
	// DDADR points to the next line
	if ( desc == 0 ) return cam->sens->h;
	--desc; // Now points to the currently active descriptor
	// There are 1+vskip DMA descriptors before line 0,
	return (desc - 1 - cam->sens->vskip) / 2;
}

void handle_data(struct cfcam_dev *cam, int line) {
	struct cfcam_sensor *sens = cam->sens;
	int l_size = sens->w*sens->bpp;
	for ( int l=cam->prev_line; l!=line && l<sens->h; ++l ) {
		if ( l>=0 && l<sens->h ) {
			unsigned long offs = cam->lines[l] - cam->dma_buf;
			dma_sync_single_for_cpu(NULL, cam->h_dma_buf+offs,
						l_size, DMA_FROM_DEVICE);
		}
		if ( l < cam->tfm->history ) continue;
		int nw = cam->tfm->transform
			(cam->tfm, cam->current_frame, l,
			 &cam->lines[l-cam->tfm->history], cam->dst);
		if ( nw > cam->remain ) panic("Bug in pxa_cfcam transform");
		cam->dst += nw;
		cam->remain -= nw;
		//printk(".");
	}
	//printk("(%d)", line);
	if ( line >= sens->h || line < cam->prev_line ) {
		// VSYNC occured
		//printk("complete\n");
		//printk("\n");
		if ( cam->state == MMAP ) {
			int offset = cam->tfm->max_frame_size*cam->kbuf_cap;
			cam->dst = cam->kbuf + offset;
			cam->remain = cam->tfm->max_frame_size;
		} else {
			cam->state = COMPLETE;
		}
	}
}

void handle(struct cfcam_dev *cam) {
	int line = current_line(cam);
	if ( line >= 2 && (cam->prev_line<2 || cam->prev_line>line) ) {
		++cam->current_frame;
		//printk("+");
	}
	switch ( cam->state ) {
	case IDLE:
		break;
	case SYNCING:
		if ( line >= 2 &&
		     (cam->prev_line<2 || cam->prev_line>line) ) {
			// VSYNC occured.
			cam->state = READING;
		}
		break;
	case READING:
	case MMAP:
		handle_data(cam, line);
		break;
	case COMPLETE:
		break;
	}
	cam->prev_line = line;
}

static u32 tref = 0;
static void tstamp(const char *msg) {
	u32 t = OSCR;
	printk("%d ms: %s\n", (t-tref)*10/36864, msg);
	tref = t;
}

static void dma_irq(int ch, void *data, struct pt_regs *regs) {
	struct cfcam_dev *cam = data;
	DCSR(cam->dma_chan) |= DCSR_ENDINTR;  // Ack interrupt
	switch ( cam->state ) {
	case SYNCING:
	case READING:
		// Must be in user context
		//wake_up(&cam->wait);
		break;
	case MMAP:
		handle(cam);
		break;
	default:
		break;
	}
}

#if 0
static struct page *
cfcam_vm_nopage(struct vm_area_struct *vma, unsigned long vaddr, int *type) {
	printk("** nopage\n");
	return NULL;
}

static struct vm_operations_struct cfcam_vm_ops = {
	.nopage   = cfcam_vm_nopage,
};
#endif

static int cfcam_mmap(struct file *file, struct vm_area_struct *vma) {
	struct cfcam_dev *cam = file->private_data;

	if ( cam->state == MMAP ) {
		printk("Already mmapped\n");
		return -EBUSY;
	}

	unsigned long size = PAGE_ALIGN(vma->vm_end - vma->vm_start);
	if ( size < 2 * cam->tfm->max_frame_size ) {
		printk("%s: user must mmap at least %d bytes\n",
		       NAME, 2 * cam->tfm->max_frame_size);
		return -EINVAL;
	}
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	printk("mmapping %ld %08lx-%08lx @%08lx\n",
	       size, vma->vm_start, vma->vm_end, offset);

	// For streaming we need memory that will be writable in IRQ context.
	// TBD Page align ?
	cam->kbuf = vmalloc_32(size);	
	if ( ! cam->kbuf ) return -ENOMEM;
	memset(cam->kbuf, 0, size);
	cam->kbuf_size = size;

	vma->vm_flags |= VM_IO;
	vma->vm_flags |= VM_RESERVED;
	unsigned long src = vma->vm_start;
	char *dst = cam->kbuf + offset;
	while ( size > 0 ) {
		if ( vm_insert_page(vma, src, vmalloc_to_page(dst)) )
			return -ENOMEM;
		src += PAGE_SIZE;
		dst += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	cam->kbuf_vma = vma;
//	vma->vm_ops = &cfcam_vm_ops;

	cam->kbuf_cap = 0;

	cam->prev_line = current_line(cam);
	cam->dst = cam->kbuf;
	cam->remain = cam->tfm->max_frame_size;
	cam->state = MMAP;

	// Wait for first frame
	int cur = cam->current_frame;
	while ( cam->current_frame == cur ) mdelay(1);

	return 0;
}

static ssize_t cfcam_read(struct file *file, char __user *buf,
			  size_t count, loff_t *offp) {    
	if ( ! access_ok(VERIFY_WRITE, buf, count) ) return -EFAULT;

	struct cfcam_dev *cam = file->private_data;
#if NOIRQ
	unsigned long flags;
#endif

	tstamp("reading");
	switch ( cam->state ) {
	case IDLE:
		cam->dst = buf;
		cam->remain = count;
#if NOIRQ
		local_irq_save(flags);
#endif
		cam->prev_line = current_line(cam);
		//init_waitqueue_head(&cam->wait);
		cam->state = SYNCING;		
		while ( cam->state == SYNCING ) {
			handle(cam);
		}
		tstamp("synced");
		while ( cam->state != COMPLETE ) {
			//interruptible_sleep_on(&cam->wait);
			handle(cam);
		}
		tstamp("complete");
		cam->state = IDLE;
#if NOIRQ
		local_irq_restore(flags);
#endif
		return cam->dst - buf;
	case MMAP: {
		if ( count < sizeof(u32) ) return -EINVAL;
		int cap = cam->kbuf_cap;
		cam->kbuf_cap = 1 - cap;
		int cur = cam->current_frame;
		//init_waitqueue_head(&cam->wait);
		int maxd = 1000;
		while ( cam->current_frame == cur && maxd-- ) {
			mdelay(1);
			//interruptible_sleep_on(&cam->wait);
		}
		int offs = cap * cam->tfm->max_frame_size;
		copy_to_user(buf, &offs, sizeof(offs));
		tstamp("swapped");
		return sizeof(u32);
	}
	default:
		return -EIO;
	}
}

//////////////////////////////////////////////////////////////////////

inline void scl0() { GPCR(gpio_scl) = GPIO_bit(gpio_scl); }
inline void scl1() { GPSR(gpio_scl) = GPIO_bit(gpio_scl); }
inline void sda0() { GPCR(gpio_sda) = GPIO_bit(gpio_sda); }
inline void sda1() { GPSR(gpio_sda) = GPIO_bit(gpio_sda); }

static void period(struct cfcam_sensor *sens) {
	for ( int i=0; i<10; ++i ) (void)*sens->mem;
	sens->i2c_skipped += 10;
	udelay(10);
}

void i2c_stop(struct cfcam_sensor *sens) {
	sda0(); pxa_gpio_mode(gpio_sda | GPIO_OUT); period(sens);
	scl1(); period(sens);
	pxa_gpio_mode(gpio_sda | GPIO_IN); period(sens);
}
void i2c_start(struct cfcam_sensor *sens) {
	sda1(); pxa_gpio_mode(gpio_sda | GPIO_OUT); period(sens);
	scl1(); period(sens);
	sda0();	period(sens); scl0(); period(sens);
}
void i2c_send(struct cfcam_sensor *sens, u8 v) {
	pxa_gpio_mode(gpio_sda | GPIO_OUT);
	for ( int i=7; i>=0; --i ) {
		if ( v & (1<<i) ) sda1(); else sda0();
		period(sens);
		scl1(); period(sens); scl0(); period(sens);
	}
	pxa_gpio_mode(gpio_sda | GPIO_IN); scl1(); period(sens);
	if ( GPLR(gpio_sda) & GPIO_bit(gpio_sda) )
		printk("Ack failed\n");
	scl0(); period(sens);
}

u8 i2c_recv(struct cfcam_sensor *sens) {
	u8 v = 0;
	pxa_gpio_mode(gpio_sda | GPIO_IN);
	for ( int i=0; i<8; ++i ) {
		scl1();
		period(sens);
		v <<= 1;
		if ( GPLR(gpio_sda) & GPIO_bit(gpio_sda) ) v |= 1;
		scl0();
		period(sens);
	}
	// Ack
	sda1(); pxa_gpio_mode(gpio_sda | GPIO_OUT); period(sens);
	scl1(); period(sens);
	scl0(); period(sens);
	return v;
}
u8 i2c_read(struct cfcam_sensor *sens, u8 reg) {
	i2c_start(sens);
	i2c_send(sens, sens->i2c_addr);
	i2c_send(sens, reg);
	i2c_start(sens);
	i2c_send(sens, sens->i2c_addr|1);
	u8 val = i2c_recv(sens);
	i2c_stop(sens);
	return val;
}
void i2c_write(struct cfcam_sensor *sens, u8 reg, u8 val) {
	i2c_start(sens);
	i2c_send(sens, sens->i2c_addr);
	i2c_send(sens, reg);
	i2c_send(sens, val);
	i2c_stop(sens);
}
void i2c_init(struct cfcam_sensor *sens) {
	pxa_gpio_mode(gpio_scl | GPIO_OUT);
	char *cfg = sens->i2c_cfg;
	while ( *cfg ) {
		char *colon;
		u8 reg = simple_strtoul(cfg, &colon, 0);
		if ( *colon != ':' ) {
			printk("Invalid I2C config string\n");
			break;
		}
		u8 val = simple_strtoul(colon+1, &cfg, 0);
		if ( *cfg == ',' )
			++cfg;
		else if ( *cfg ) {
			printk("Invalid I2C config string\n");
			break;
		}
		//printk("I2C 0x%x : 0x%x\n", reg, val);
		i2c_write(sens, reg, val);
	}

	pxa_gpio_mode(gpio_scl | GPIO_IN);
}

//////////////////////////////////////////////////////////////////////

static int setup_dev(struct cfcam_dev *cam) {
	int ret;

	init_MUTEX(&cam->opened);
 	
	if ( sensor >= ARRAY_SIZE(sensors) ) {
		ret = -EINVAL;
		goto fail1;
	}
	struct cfcam_sensor *sens = &sensors[sensor];
	printk("Sensor configuration #%d: %s\n", sensor, sens->name);
	
	ret = pxa_request_dma(NAME, DMA_PRIO_LOW, dma_irq, cam); // TBD
	if ( ret < 0 ) goto fail1;
	cam->dma_chan = ret;
	
	cam->sens = sens;

	// TBD Must be more than IRQ latency and less than 2M.
	// Should also be a submultiple of total size so that
	// data is not overwritten too early after the last line.
	cam->dma_size = sens->w * sens->h * sens->bpp;
	while ( cam->dma_size > 1500000 && (cam->dma_size&1) == 0 )
		cam->dma_size /= 2;

	// It would be simpler to get uncached memory with
	// dma_alloc_coherent, but then the color-conversion
	// code would eat a lot of bandwidth on the memory bus.
	// So instead we use normal cacheable memory and invalidate
	// the cache before reading.
	cam->dma_buf = kmalloc(cam->dma_size, GFP_KERNEL|GFP_DMA);
	if ( ! cam->dma_buf ) { ret = -ENOMEM; goto fail2; }
	cam->h_dma_buf = dma_map_single(0, cam->dma_buf, cam->dma_size,
					DMA_FROM_DEVICE);
	if ( !cam->h_dma_buf || !~cam->h_dma_buf ){ ret=-ENOMEM; goto fail3; }
	memset(cam->dma_buf, 0, cam->dma_size); // cache
	int ndescs = 1 + sens->vskip + 2*sens->h;
	cam->descs_size = sizeof(struct pxa_dma_desc) * ndescs;
	cam->descs = kmalloc(cam->descs_size, GFP_KERNEL|GFP_DMA);
	if ( ! cam->descs ) { ret = -ENOMEM; goto fail4; }

	cam->h_descs = dma_map_single(0, cam->descs, cam->descs_size,
				      DMA_TO_DEVICE);
	if ( !cam->h_descs || !~cam->h_descs ) {
		ret = -ENOMEM; goto fail5;
	}

	cam->lines = kmalloc(sizeof(char*)*sens->h, GFP_KERNEL);
	if ( ! cam->lines ) { ret = -ENOMEM; goto fail6; }
	
	int line_size = sens->w * sens->bpp;
	int lines_in_buffer = cam->dma_size / line_size;;

	u32 cmd = DCMD_BURST8 | DCMD_WIDTH2 | DCMD_ENDIRQEN;

	int l, bl;
	u32 bytes_per_frame = 0;
	struct pxa_dma_desc *pd = cam->descs;
	struct pxa_dma_desc *h_pd = (void*)cam->h_descs;
	// Skip
	pd->dsadr = addr;
	pd->dtadr = cam->h_dma_buf;
	pd->dcmd = cmd | sens->skip;
	bytes_per_frame += sens->skip;
	pd++->ddadr = (u32)++h_pd;
	// vskip
	for ( l=0; l<sens->vskip; ++l ) {
		pd->dsadr = addr;
		pd->dtadr = cam->h_dma_buf;
		pd->dcmd = cmd | (sens->hskip+line_size);
		bytes_per_frame += (sens->hskip+line_size);
		pd++->ddadr = (u32)++h_pd;
	}		
	for ( l=0; l<sens->h; ++l ) {
		bl = l % lines_in_buffer;
		// hskip
		pd->dsadr = addr;
		pd->dtadr = cam->h_dma_buf + line_size*bl;
		pd->dcmd = cmd | sens->hskip;
		bytes_per_frame += sens->hskip;
		pd++->ddadr = (u32)++h_pd;
		// line data
		cam->lines[l] = cam->dma_buf + line_size*bl;
		pd->dsadr = addr;
		pd->dtadr = cam->h_dma_buf + line_size*bl;
		pd->dcmd = cmd | DCMD_INCTRGADDR | line_size;
		bytes_per_frame += line_size;
		pd++->ddadr = (u32)++h_pd;
	}
	// Last DMA descriptor loops back to first.
	(pd-1)->ddadr = cam->h_descs;

	// Make sure the DMAC sees the newly written descriptors.
	dma_sync_single_for_cpu(NULL, cam->h_descs, cam->descs_size,
				DMA_TO_DEVICE);

 	sens->mem = ioremap_nocache(addr, 4096);

#if 0
	// Chip init and timing
	u32 t0 = OSCR;
	register volatile u32 *xx = sens->mem;
	for ( register int n=10000; n--; ) (void)*xx;
	u32 t1 = OSCR;
	printk("%d clocks in %d ticks (%d/s)\n",
	       10000, t1-t0, 10000*36468/((t1-t0)/100));
#endif

	printk("Frame size at DMA level = %d clocks\n", bytes_per_frame/2);

	return 0;

fail6:  dma_unmap_single(0, cam->h_descs, cam->descs_size, DMA_TO_DEVICE);
fail5:  kfree(cam->descs);
fail4:  dma_unmap_single(0, cam->h_dma_buf, cam->dma_size, DMA_FROM_DEVICE);
fail3:  kfree(cam->dma_buf);
fail2:  pxa_free_dma(cam->dma_chan);
fail1:  return ret;
}

static void release_dev(struct cfcam_dev *cam) {
	iounmap((void*)cam->sens->mem);
	kfree(cam->lines);
	dma_unmap_single(0, cam->h_descs, cam->descs_size, DMA_TO_DEVICE);
	kfree(cam->descs);
	dma_unmap_single(0, cam->h_dma_buf, cam->dma_size, DMA_FROM_DEVICE);
	kfree(cam->dma_buf);
	pxa_free_dma(cam->dma_chan);
}

//////////////////////////////////////////////////////////////////////

static int start_cam(struct cfcam_dev *cam) {
	struct cfcam_sensor *sens = cam->sens;

	// Reset on
	GPCR(gpio_reset) = GPIO_bit(gpio_reset);
	udelay(1);
 	// Clock the reset signal into the sensor
	for ( int i=0; i<20; ++i ) (void)*sens->mem;
	// Reset off
	GPSR(gpio_reset) = GPIO_bit(gpio_reset);
	mdelay(1);

	// Optional I2C init
	sens->i2c_skipped = 0;
	if ( sens->i2c_cfg ) i2c_init(sens);

	// Align to first frame.
	int preskip = sens->preskip - sens->i2c_skipped;
	if ( preskip < 0 ) printk("Can't clock to start of frame\n");
	while ( preskip-- > 0 ) (void)*sens->mem;

	// Ready to start DMA
	DDADR(cam->dma_chan) = cam->h_descs;
	DCSR(cam->dma_chan) |= DCSR_RUN;

	printk("Camera running\n");
	cam->state = IDLE;
	cam->current_frame = 0;

	// Wait for first frame
	mdelay(100);

	return 0;
}

static void stop_cam(struct cfcam_dev *cam) {
	DCSR(cam->dma_chan) &= ~DCSR_RUN;
	mdelay(10);
	printk("Camera stopped\n");
}

//////////////////////////////////////////////////////////////////////

static pxa_cfcam_createfn transform_createfn = NULL;

int pxa_cfcam_set_transform(pxa_cfcam_createfn fn) {
	if ( transform_createfn ) return -EBUSY;
	transform_createfn = fn;
	return 0;
}
void pxa_cfcam_unset_transform() {
	transform_createfn = NULL;
}

EXPORT_SYMBOL(pxa_cfcam_set_transform);
EXPORT_SYMBOL(pxa_cfcam_unset_transform);

#define NCAMS 1
static struct cfcam_dev cams[NCAMS];

static int cfcam_open(struct inode *inode, struct file *file) {
	int ret;
	printk("open...\n");

	if ( ! transform_createfn ) return -ENODEV;

	int index = MINOR(inode->i_rdev);
	if ( index<0 || index>=NCAMS ) return -ENODEV;
	struct cfcam_dev *cam = &cams[index];

	if ( down_interruptible(&cam->opened) ) return -EINTR;

	cam->tfm = transform_createfn
		(cam->sens->w, cam->sens->h, 9, cam->sens->bpp);
	if ( ! cam->tfm ) { ret = -ENODEV; goto fail1; }

	file->private_data = cam;
	cam->kbuf = NULL;
	printk("start...\n");
	start_cam(cam);
	return 0;

fail1:  up(&cam->opened);
	return ret;
}

static int cfcam_release(struct inode *inode, struct file *file) {
	struct cfcam_dev *cam = file->private_data;
	stop_cam(cam);
	if ( cam->kbuf ) vfree(cam->kbuf);
	cam->tfm->release(cam->tfm);
	up(&cam->opened);
	return 0;
}

static struct file_operations cfcam_fops = {
	.owner   = THIS_MODULE,
	.open    = cfcam_open,
	.mmap    = cfcam_mmap,
	.read    = cfcam_read,
	.release = cfcam_release,
};

/**************** Module ****************/

static struct class *class;

static int __init pxa_cfcam_init() {
	printk(NAME ": init...\n");
	printk("GPIO%d = reset\n", gpio_reset);
	printk("GPIO%d = powerdown\n", gpio_powerdown);
	printk("GPIO%d = sda\n", gpio_sda);
	printk("GPIO%d = scl\n", gpio_scl);

	int ret = register_chrdev(dev_major, NAME, &cfcam_fops);
	if ( ret < 0 ) goto fail1;
	if ( ! dev_major ) dev_major = ret;

	class = class_create(THIS_MODULE, NAME);
	if ( ! class ) { ret = -EBUSY; goto fail2; }

//	MCIO(0) = (5<<14) | (5<<7) | 5;
	MCIO(0) = (0<<14) | (0<<7) | 0;

	// Enable PCMCIA signals required to drive bus transceivers
	pxa_gpio_mode(GPIO54_pSKTSEL_MD);
	MECR = MECR_CIT; // Side effect: DMA always 16-bit access

	pxa_gpio_mode(GPIO50_nPIOR_MD);
	pxa_gpio_mode(GPIO51_nPIOW_MD);

	// Power/reset pins
	pxa_gpio_mode(gpio_reset | GPIO_OUT);
	GPSR(gpio_powerdown) = GPIO_bit(gpio_powerdown);
	pxa_gpio_mode(gpio_powerdown | GPIO_OUT);

	int i;
	for ( i=0; i<NCAMS; ++i ) {
		ret = setup_dev(&cams[i]);
		if ( ret ) {
			while ( i-- ) release_dev(&cams[i]);
			goto fail3;
		}
		class_device_create(class, NULL, MKDEV(dev_major,i),
				    NULL, NAME"%d", i);
	}

	return 0;

fail3:	class_destroy(class);
fail2:  unregister_chrdev(dev_major, NAME);
fail1:  return ret;
}

static void __exit pxa_cfcam_exit() {
	printk("%s exit...\n", NAME);
	class_destroy(class);
	int i;
	for ( i=0; i<NCAMS; ++i )
		release_dev(&cams[i]);
	unregister_chrdev(dev_major, NAME);
}

module_init(pxa_cfcam_init);
module_exit(pxa_cfcam_exit);
