/* -*- linux-c -*-
 *
 * Bayer-to-RGB transform for pxa_cfcam
 * Copyright (c) 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/module.h>

#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/leds.h>
#include <asm/irq.h>
#include <asm/fiq.h>
#include <asm/mach/irq.h>
#include <asm/mach/time.h>
#include <asm/arch/pxa-regs.h>
#include <asm/cacheflush.h>

#include <asm/arch/dma.h>

#include <linux/dma-mapping.h>

#include "pxa_cfcam.h"

#define NAME "pxa_cfcam_pnm"

static int bpp = 3;
module_param(bpp, int, 0400);
MODULE_PARM_DESC(bpp, "Output image bytes per pixel (1=Y, 3=RGB)");

static int pnm_header = 1;
module_param(pnm_header, int, 0400);
MODULE_PARM_DESC(pnm_header, "Output a PNM header (default: y)");

#if 1
static int raw = 0;
module_param(raw, int, 0400);
MODULE_PARM_DESC(raw, "Output raw data (requires bpp=1)");
#endif

//////////////////////////////////////////////////////////////////////

// The conversion functions (could use some asm optimization)

static void grbg932_to_y8(u32 *srct, u32 *src, u32 *srcb, u8 *dst, int w) {
	for ( ; w; w-=2 ) {
		++src; ++srct; ++srcb;
		*dst++ = ((srct[-1]+srct[0]+srct[+1]+
			   src[-1]+4*src[0]+src[1]+
			   srcb[-1]+srcb[0]+srcb[+1]) & 65535) / 24;
		++src; ++srct; ++srcb;
		*dst++ = ((srct[0]+
			   src[-1]+2*src[0]+src[1]+
			   srcb[0]) & 65535) / 12;
	}
}

static void bggr932_to_y8(u32 *srct, u32 *src, u32 *srcb, u8 *dst, int w) {
	for ( ; w; w-=2 ) {
		++src; ++srct; ++srcb;
		*dst++ = ((srct[0]+
			   src[-1]+2*src[0]+src[1]+
			   srcb[0]) & 65535) / 12;
		++src; ++srct; ++srcb;
		*dst++ = ((srct[-1]+srct[0]+srct[+1]+
			   src[-1]+4*src[0]+src[1]+
			   srcb[-1]+srcb[0]+srcb[+1]) & 65535) / 24;
	}
}
	
static void grbg932_to_rgb8(u32 *srct, u32 *src, u32 *srcb, u8 *dst, int w) {
	for ( ; w; w-=2 ) {
		++src; ++srct; ++srcb;
		*dst++ = src[0]/2;
		*dst++ = (srct[0]+srcb[0]+src[-1]+src[+1]) / 8;
		*dst++ = (srct[-1]+srct[+1]+srcb[-1]+srcb[+1]) / 8;
		++src; ++srct; ++srcb;
		*dst++ = (src[-1]+src[+1]) / 4;
		*dst++ = src[0]/2;
		*dst++ = (srct[0]+srcb[0]) / 4;
	}
}
	
static void bggr932_to_rgb8(u32 *srct, u32 *src, u32 *srcb, u8 *dst, int w) {
	for ( ; w; w-=2 ) {
		++src; ++srct; ++srcb;
		*dst++ = (srct[0]+srcb[0]) / 4;
		*dst++ = src[0]/2;
		*dst++ = (src[-1]+src[+1]) / 4;
		++src; ++srct; ++srcb;
		*dst++ = (srct[-1]+srct[+1]+srcb[-1]+srcb[+1]) / 8;
		*dst++ = (srct[0]+srcb[0]+src[-1]+src[+1]) / 8;
		*dst++ = src[0]/2;
	}
}
	
//////////////////////////////////////////////////////////////////////

// pxa_cfcam interface

struct pxa_cfcam_pnm_transform {
	int out_w, out_h;
	int in_bpp;
	struct pxa_cfcam_transform tfm;
};

static int transform(struct pxa_cfcam_transform *tfm,
		     int frame, int line, char **lines, char *dst) {
	struct pxa_cfcam_pnm_transform *t = tfm->priv_data;
	int ret = 0;
	
	if ( line==2 && pnm_header ) {
		char *p = (bpp==1) ? "P5" : "P6";
		sprintf(dst, "%s\n%d %d\n255\n", p, t->out_w, t->out_h);
		ret = strlen(dst);
		dst += ret;
	}

#if 1
	if ( raw ) {
		u16 *s = (u16*)lines[1];
		for ( int n=t->out_w; n--; )
			*dst++ = *++s >> 1;
		ret += t->out_w;
	} else
#endif
	switch ( bpp ) {
	case 1:
		if ( ! (line&1) ) // BGBGBG GRGRGR BGBGBG
			grbg932_to_y8((u32*)lines[0], (u32*)lines[1],
				      (u32*)lines[2], dst, t->out_w);
		else // GRGRGR BGBGBG GRGRGR
			bggr932_to_y8((u32*)lines[0], (u32*)lines[1],
				      (u32*)lines[2], dst, t->out_w);
		ret += t->out_w;
		break;
	case 3:
		if ( ! (line&1) ) // BGBGBG GRGRGR BGBGBG
			grbg932_to_rgb8((u32*)lines[0], (u32*)lines[1],
					(u32*)lines[2], dst, t->out_w);
		else // GRGRGR BGBGBG
			bggr932_to_rgb8((u32*)lines[0], (u32*)lines[1],
					(u32*)lines[2], dst, t->out_w);
		ret += t->out_w*3;
		break;
	}
	return ret;
}

static void pxa_cfcam_pnm_free(struct pxa_cfcam_transform *tfm) {
	struct pxa_cfcam_pnm_transform *t = tfm->priv_data;
	kfree(t);
	//module_put(THIS_MODULE);
}

#define PNM_HEADER_SIZE (2+1 + 5+1+5+1+ 3+1) // "P6\nWWWWW HHHHH\n255\n"

static struct pxa_cfcam_transform *
pxa_cfcam_pnm_alloc(int in_w, int in_h, int in_bits, int in_bpp) {
	if ( in_bits != 9 ) {
		printk("%s: requires 9 bits per input pixel.\n", NAME);
		return NULL;
	}
	if ( raw && (in_bpp!=2 || bpp!=1) ) {
		printk("%s: raw mode requires in_bpp=2 and bpp=1.\n", NAME);
		return NULL;
	}
	if ( !raw && in_bpp!=2 && in_bpp!=4 ) {
		printk("%s: requires 2 or 4 bytes per input pixel.\n", NAME);
		return NULL;
	}
	if ( !raw && in_bpp!=4 ) {
		printk("%s: Bayer requires 4 bytes per input pixel.\n", NAME);
		return NULL;
	}
	if ( in_w<3 || in_w>10000 || in_h<3 || in_h>10000 ) {
		printk("%s: unsupported input size.\n", NAME);
		return NULL;
	}
	struct pxa_cfcam_pnm_transform *t =
		kmalloc(sizeof(struct pxa_cfcam_pnm_transform), GFP_KERNEL);
	if ( ! t ) return NULL;

	t->out_w = in_w - 2;
	t->out_h = in_h - 2;
	t->in_bpp = in_bpp;
	t->tfm.max_frame_size = t->out_w * t->out_h * bpp;
	if ( pnm_header ) t->tfm.max_frame_size += PNM_HEADER_SIZE;
	t->tfm.history = 2;
	t->tfm.priv_data = t;
	t->tfm.transform = transform;
	t->tfm.release = pxa_cfcam_pnm_free;

	// Forbid unloading while pxa_cfcam is referencing us.
        // TBD __module_get(THIS_MODULE);

	return &t->tfm;
}

//////////////////////////////////////////////////////////////////////
  
// Module interface

static int __init pxa_cfcam_pnm_init() {
	return pxa_cfcam_set_transform(pxa_cfcam_pnm_alloc);
}

static void __exit pxa_cfcam_pnm_exit() {
	printk("%s exit...\n", NAME);
	pxa_cfcam_unset_transform();
}

MODULE_LICENSE("GPL");
module_init(pxa_cfcam_pnm_init);
module_exit(pxa_cfcam_pnm_exit);
