/* -*- linux-c -*-
 *
 * PWM/MPWM/DCM signal decoder for PXA255 GPIO.
 * Copyright (c) 2005, 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/device.h>

#include <asm/types.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/mach/time.h>
#include <asm/arch/pxa-regs.h>
#include <asm/cacheflush.h>

#include "pxa_rtsched.h"

/* Hardware constants */
#define MIN_GPIO 2
#define MAX_GPIO 84

#define NAME "ipwm"

static int dev_major = 0;
module_param(dev_major, int, 0400);
MODULE_PARM_DESC(dev_major, "Device major number (default: auto-alloc)");

#define MAX_PINS 4   /* Max # of pins (1-16). */
#define MAX_CHANS 8  /* Max # of channels per pin (1-15) */
#define PIN_NUM(dev) (MINOR(dev)>>4)
#define CHAN_NUM(dev) (MINOR(dev)&15)

/* Counter mode can be disabled here to make the FIQ handler faster */
#define COUNTER_MODE 1

#if USE_CACHE_LOCKING
#include "pxa_cache_lock.h"
#define ILOCK __attribute__((section(".lockable.text")))
void ILOCK i_lock_begin(void) { }
#else
#define ILOCK
#endif

typedef u16 val_t;
#if COUNTER_MODE
typedef u32 counter_t;
#endif

enum mode { PWM, PPM, DCM, COUNTER,
	    MODE_MAX=COUNTER } __attribute__((packed));

struct pin {
	char gpio;
	enum mode mode;
	unsigned char nchans;
	pxa_rtc_t tmin, tmax, tsync, timeout;
};

struct event { pxa_rtc_t ts; u32 val; };

#define MAX_EV (MAX_PINS*(MAX_CHANS+1)*2*4)

static struct {
	int npins;
	struct pin pins[MAX_PINS];	
	struct event events[MAX_EV];
	int next_ev;
	volatile u32 *gedr;
	volatile u32 *gplr;
#if COUNTER_MODE
	u32 counter_mask;
	volatile counter_t counters[32];
	signed char firstbit[16];
#endif
} lock;

/* Module API */

MODULE_DESCRIPTION("PWM/PPM/DCM decoder and pulse counter for PXA255 GPIO");
MODULE_AUTHOR("pascal@pabr.org");
MODULE_LICENSE("GPL");

static int gpio[MAX_PINS] = { [0 ... MAX_PINS-1] = -1 };
module_param_array(gpio, int, NULL, 0400);
MODULE_PARM_DESC(gpio, "GPIO pins");

static int mode[MAX_PINS] = { [0 ... MAX_PINS-1] = PWM };
module_param_array(mode, int, NULL, 0400);
MODULE_PARM_DESC(mode, "0:PWM 1:PPM 2:DCM 3:Counter");

static int nchans[MAX_PINS] = { [0 ... MAX_PINS-1] = 1 };
module_param_array(nchans, int, NULL, 0400);
MODULE_PARM_DESC(nchans, "Number of channels");

static int tmin[MAX_PINS] = { [0 ... MAX_PINS-1] = 1000 };
module_param_array(tmin, int, NULL, 0400);
MODULE_PARM_DESC(tmin, "Width for 0% / Min period (microseconds)");

static int tmax[MAX_PINS] = { [0 ... MAX_PINS-1] = 2000 };
module_param_array(tmax, int, NULL, 0400);
MODULE_PARM_DESC(tmax, "Width for 100% / Max period (microseconds)");

static int tsync[MAX_PINS] = { [0 ... MAX_PINS-1] = 4000 };
module_param_array(tsync, int, NULL, 0400);
MODULE_PARM_DESC(tsync, "Min time between pulse trains (microseconds)");

static int timeout[MAX_PINS] = { [0 ... MAX_PINS-1] = 100000 };
module_param_array(timeout, int, NULL, 0400);
MODULE_PARM_DESC(timeout, "Max time between pulse trains (microseconds)");

/**********************************************************************/
/*                    Interrupt handler                               */
/**********************************************************************/

void ILOCK rt_handler(pxa_rtc_t t) {
#if COUNTER_MODE
	u32 v = *lock.gedr;
#endif
	*lock.gedr = -1; /* Clear interrupt */
	if ( lock.next_ev == 0 ) lock.next_ev = MAX_EV;
	--lock.next_ev;
	lock.events[lock.next_ev].ts = t;
	lock.events[lock.next_ev].val = *lock.gplr;
#if COUNTER_MODE
	v &= lock.counter_mask;
	if ( v ) {
		for ( int base=0; base<32; base+=4,v>>=4 ) {
			int b;
			while ( (b=lock.firstbit[v&15]) >= 0 ) {
				++lock.counters[base+b];
				v ^= 1<<b;
			}
		}
	}
#endif
}

#if USE_CACHE_LOCKING
void ILOCK i_lock_end(void) { }
#endif

/**********************************************************************/
/*                    PWM/PPM/DCM decoding                            */
/**********************************************************************/

#define NEXT \
  { t = lock.events[e].ts;                      \
    if ( (s32)(t-tmin) < 0 ) return -ETIMEDOUT;	\
    if ( ++e == MAX_EV ) e = 0;                 \
    if ( e == emax ) return -ETIMEDOUT;         \
  }
#define NEXT_0   while (    lock.events[e].val & mask  ) NEXT
#define NEXT_1   while ( ! (lock.events[e].val & mask) ) NEXT

inline val_t convert(struct pin *pin, int dt) {
	dt -= pin->tmin;
	if ( dt < 0 ) dt = 0;
	dt = (dt<<16) / (pin->tmax-pin->tmin);
	if ( dt >= 65536 ) dt = 65535;
	return dt;
}

int decode(struct pin *pin, val_t *buf) {
	u32 mask = 1 << (pin->gpio&31);
	int c;
	int e = lock.next_ev;
	int emax = ( e + MAX_EV/2 ) % MAX_EV;
	pxa_rtc_t tmin = OSCR - pin->timeout;
	pxa_rtc_t t = 0; /* Avoid irrelevant GCC warning */
	pxa_rtc_t t0, t1;
	int dt;

	switch ( pin->mode ) {
	case PWM:
		NEXT_0; NEXT_1; t1 = t; NEXT_0; t0 = t;
		NEXT_1; if ( t0-t < pin->tsync ) return -EIO;
		*buf = convert(pin, t1-t0);
		return 0;
	case PPM:
		NEXT_0;
		for ( c=pin->nchans; c--; ) {
			t1 = t;
			NEXT_1;
			NEXT_0;
			dt = t1 - t;
			if ( dt > pin->tsync ) { c = pin->nchans; continue; }
			buf[c] = convert(pin, dt);
			if ( ! c ) {
				t1 = t;
				NEXT_1;
				if ( t1 - t < pin->tsync ) c = pin->nchans;
			}
		}
		return 0;
	case DCM:
		NEXT_1; NEXT_0;
		t1 = t; NEXT_1; t0 = t; NEXT_0;
		dt = t1 - t;
		if ( dt<pin->tmin || dt>pin->tmax ) return -EIO;
		t1 = t0 - t;
		*buf = (t1<<16) / dt;
		return 0;
	default:
		return -EFAULT;
	}
}

/**********************************************************************/
/*                         File operations                            */
/**********************************************************************/

/* Pin devices for PWM/PPM/DCM mode. */

static ssize_t ipwm_pin_read(struct file *file, char *buf,
			     size_t count, loff_t *offp) {    
	struct pin *pin = (struct pin*)file->private_data;
	int size = pin->nchans * sizeof(val_t);
	val_t buf1[pin->nchans];
	int ret;
	if ( count < size ) return -EINVAL;
	ret = decode(pin, buf1);
	if ( ret ) return ret;
	if ( copy_to_user(buf, buf1, size) ) return -EFAULT;
	return size;
}

static int ipwm_pin_release(struct inode *inode, struct file *file) {
	return 0;
}

static struct file_operations ipwm_pin_fops = {
	.owner   = THIS_MODULE,
	.read    = ipwm_pin_read,
	.release = ipwm_pin_release,
};

/* Channel devices for PPM mode only. */

static ssize_t ipwm_chan_read(struct file *file, char *buf,
			      size_t count, loff_t *offp) {    
	int p = (u32)file->private_data >> 8;
	struct pin *pin = &lock.pins[p];
	int c = (u32)file->private_data & 255;
	val_t vals[MAX_CHANS];
	int ret;
	if ( count < sizeof(val_t) ) return -EINVAL;
	ret = decode(pin, vals);
	if ( ret ) return ret;
	if ( put_user(vals[c], (val_t*)buf) ) return -EFAULT;
	return sizeof(val_t);
}

static int ipwm_chan_release(struct inode *inode, struct file *file) {
	return 0;
}

static struct file_operations ipwm_chan_fops = {
	.owner   = THIS_MODULE,
	.read    = ipwm_chan_read,
	.release = ipwm_chan_release,
};

/* Devices for COUNTER mode only. */

#if COUNTER_MODE

static ssize_t ipwm_counter_read(struct file *file, char *buf,
				 size_t count, loff_t *offp) {
	struct pin *pin = (struct pin*)file->private_data;
	if ( count < sizeof(counter_t) ) return -EINVAL;
	counter_t v = lock.counters[pin->gpio&31];
	if ( ! pin->tmin ) /* First call never blocks */
		pin->tmin = 1;
	else /* Subsequent calls wait for a new event */
		while ( (u32)v == (u32)pin->tsync ) {
			msleep_interruptible
				(pin->timeout * 10/(RTSCHED_FREQ/100));
			if ( signal_pending(current) ) return -EINTR;
			v = lock.counters[pin->gpio&31];
		}
	if ( put_user(v, (counter_t*)buf) ) return -EFAULT;
	pin->tsync = v; // Remember the last value returned
	return sizeof(counter_t);
}

static ssize_t ipwm_counter_write(struct file *file, const char *buf,
				  size_t count, loff_t *offp) {
	struct pin *pin = (struct pin*)file->private_data;
	counter_t c;
	if ( count < sizeof(counter_t) ) return -EINVAL;
	if ( get_user(c, (counter_t*)buf) ) return -EFAULT;
	lock.counters[pin->gpio&31] = c;
	return sizeof(counter_t);
}

static int ipwm_counter_release(struct inode *inode, struct file *file) {
	return 0;
}

static struct file_operations ipwm_counter_fops = {
	.owner   = THIS_MODULE,
	.read    = ipwm_counter_read,
	.write   = ipwm_counter_write,
	.release = ipwm_counter_release,
};

#endif /* COUNTER_MODE */


static int ipwm_open(struct inode *inode, struct file *file) {
	int p = PIN_NUM(inode->i_rdev);
	int c = CHAN_NUM(inode->i_rdev);
	struct pin *pin;
	if ( p<0 || p>=MAX_PINS ) return -ENODEV;
	pin = &lock.pins[p];
	if ( c == 15 ) {
		file->f_op = &ipwm_pin_fops;
#if COUNTER_MODE
		if ( pin->mode == COUNTER ) {
			file->f_op = &ipwm_counter_fops;
			pin->tmin = 0;
		}
#endif
		file->private_data = pin;
	} else {
		if ( c<0 || c>=pin->nchans || pin->mode == COUNTER )
			return -ENODEV;
		file->f_op = &ipwm_chan_fops;
		file->private_data = (void*)((p<<8) | c);
	}
	return 0;
}

static struct file_operations ipwm_fops = {
	.owner = THIS_MODULE,
	.open  = ipwm_open,
};

/**********************************************************************/
/*                             Module                                 */
/**********************************************************************/

static struct class *class;

static int __init pxa_ipwm_init(void) {
	int ret, p;
	int gpreg, gpmask;
	u32 re[3] = { 0, 0, 0 }; /* Interrupt mask for rising edges */
	u32 fe[3] = { 0, 0, 0 }; /* Interrupt mask for falling edges */

	memset(&lock, 0, sizeof(lock));

	signed char fb[16] = { -1,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0 };
	memcpy(lock.firstbit, fb, sizeof(lock.firstbit));

	/* Check config. */

	for ( p=0; p<MAX_PINS && gpio[p]>=0; ++p ) {
		struct pin *pin = &lock.pins[p];
		volatile u32 *new_gedr, *new_gplr;
		pin->gpio = gpio[p];
		pin->nchans = nchans[p];
		pin->mode = mode[p];
		pin->tmin = us2clocks(tmin[p]);
		pin->tmax = us2clocks(tmax[p]);
		pin->tsync = us2clocks(tsync[p]);
		pin->timeout = us2clocks(timeout[p]);
		printk(NAME "%d on gpio %d: "
		       "nchans=%d mode=%d tmin=%d tmax=%d "
		       "tsync=%d timeout=%d\n",
		       p, pin->gpio, pin->nchans, pin->mode,
		       pin->tmin, pin->tmax, pin->tsync, pin->timeout);
		if ( pin->gpio < MIN_GPIO || pin->gpio > MAX_GPIO ||
		     pin->nchans < 1 || pin->nchans > MAX_CHANS ||
		     pin->tmin <= 0 || pin->tmax <= pin->tmin ||
		     pin->mode > MODE_MAX ||
		     (pin->mode == PWM && pin->nchans != 1) ||
		     (pin->mode == PPM && pin->tsync <= pin->tmax) ||
		     (pin->mode == COUNTER && pin->nchans != 1) ) {
			printk(KERN_WARNING NAME
			       ": Invalid configuration.\n");
			ret = -EINVAL;
			goto fail1;
		}
		gpreg = pin->gpio >> 5;
		gpmask = 1 << (pin->gpio&31);
		new_gedr = &GEDR0 + gpreg;
		new_gplr = &GPLR0 + gpreg;
		if ( lock.gedr && new_gedr!=lock.gedr ) {
			printk(KERN_WARNING NAME
			       ": Pins must be in the same 32 bit word.\n");
			ret = -EINVAL;
			goto fail1;
		}
		pxa_gpio_mode(pin->gpio | GPIO_IN);
		lock.gedr = new_gedr;
		lock.gplr = new_gplr;
		re[gpreg] |= gpmask;
		if ( pin->mode != COUNTER ) fe[gpreg] |= gpmask;
#if COUNTER_MODE
		if ( pin->mode == COUNTER ) lock.counter_mask |= gpmask;
#endif
	}
	lock.npins = p;
	
	/* Register device */

	ret = register_chrdev(dev_major, NAME, &ipwm_fops);
	if ( ret < 0 ) goto fail1;
	if ( ! dev_major ) dev_major = ret;

	class = class_create(THIS_MODULE, NAME);
	if ( ! class ) { ret = -EBUSY; goto fail2; }
	for ( p=0; p<lock.npins; ++p ) {
		struct pin *pin = &lock.pins[p];
		class_device_create(class, NULL, MKDEV(dev_major,p*16+15),
				    NULL, NAME"%d", p);
		int c;
		if ( pin->mode==PPM ) {
			for ( c=0; c<pin->nchans; ++c )
				class_device_create
					(class, NULL, MKDEV(dev_major,p*16+c),
					 NULL, NAME"%d-%d", p, c);
		}
	}

#if USE_CACHE_LOCKING
	printk(NAME ": code %p-%p, data %p - %p\n",
	       i_lock_begin, i_lock_end,
	       &lock, &lock+1);
	lock_i_cache(i_lock_begin, i_lock_end);
	lock_d_cache(&lock, &lock+1);
	/* Must lock constants that GCC puts in .text. */
	lock_d_cache(i_lock_begin, i_lock_end); /* TBD Overkill */
#endif
#if USE_CACHE_LOCKING && USE_TLB_LOCKING
	lock_i_tlb(i_lock_begin, i_lock_end);
	lock_d_tlb(&lock, &lock+1);
	lock_d_tlb((void*)&GPLR0, (void*)&GPLR0+1);  // GPIO regs
        /* Must lock constants that GCC puts in .text. */
	lock_d_tlb(i_lock_begin, i_lock_end);
#endif

	ret = rtsched_set_edge_handler(rt_handler, re, fe);
	if ( ret ) {
		printk(KERN_WARNING NAME
		       ": Unable to claim handler\n");
		goto fail3;
	}
	
	printk(NAME": Initialized, %d pin(s).\n", lock.npins);
	return 0;

fail3:
#if USE_CACHE_LOCKING && USE_TLB_LOCKING
	unlock_i_tlb();
	unlock_d_tlb();
#endif
#if USE_CACHE_LOCKING
	unlock_i_cache(); /* TBD: finer-grained. */
	unlock_d_cache();
#endif
	class_destroy(class);
fail2:	unregister_chrdev(dev_major, NAME);
fail1:  printk(KERN_WARNING NAME": Aborting.\n");
	return ret;
}

static void __exit pxa_ipwm_exit(void) {
	printk(NAME ": Exiting.\n");

	class_destroy(class);
	unregister_chrdev(dev_major, NAME);

#if SAFE_RMMOD
	/* Shutdown RT before we mess with cache locks. */
	rtsched_stop();
#else
	rtsched_remove_edge_handler(rt_handler);
#endif

#if USE_CACHE_LOCKING && USE_TLB_LOCKING
	unlock_i_tlb();
	unlock_d_tlb();
#endif
#if USE_CACHE_LOCKING
	unlock_i_cache(); /* TBD: finer-grained. */
	unlock_d_cache();
#endif
}

module_init(pxa_ipwm_init);
module_exit(pxa_ipwm_exit);
