/* -*- linux-c -*-
 *
 * PWM signal generator, using pxa_rtsched.
 * Copyright (c) 2005, 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/device.h>

#include <asm/types.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/mach/time.h>
#include <asm/arch/pxa-regs.h>
#include <asm/cacheflush.h>

#include "pxa_rtsched.h"

/* Hardware constants */
#define MIN_GPIO 0
#define MAX_GPIO 84

#define NAME "opwm"

static int dev_major = 0;
module_param(dev_major, int, 0400);
MODULE_PARM_DESC(dev_major, "Device major number (default: auto-alloc)");

#define MAX_PINS 2   /* Max # of pins (1-16). */
#define MAX_CHANS 8  /* Max # of channels per pin (1-15) */
#define PIN_NUM(dev) (MINOR(dev)>>4)
#define CHAN_NUM(dev) (MINOR(dev)&15)

#if USE_CACHE_LOCKING
#include "pxa_cache_lock.h"
#define ILOCK __attribute__((section(".lockable.text")))
void ILOCK i_lock_begin(void) { }
#else
#define ILOCK
#endif

#if USE_CACHE_LOCKING && USE_TLB_LOCKING
#  define STARTFRAME_DURATION 150   // Enough for 6-channel PPM
#  define UPDOWN_DURATION 30
#else
#  define STARTFRAME_DURATION 200
#  define UPDOWN_DURATION 100
#endif

typedef u16 val_t;

enum mode { PWM, PPM, DCM, MODE_MAX=DCM } __attribute__((packed));

enum status { IN_USE=1, VALID=2 } __attribute__((packed));;

struct pin {
	unsigned char status;
	signed char gpio; /* Multiplexed output. -1 if disabled. */
	unsigned char nchans;
	enum mode mode;
	pxa_rtc_t tmin, tmax;
	struct chan {
		val_t val;
		char in_use;
		signed char gpio; /* Servo output. -1 if disabled. */
		struct pin *pin; /* TBD wasted */
		struct pxa_rttask up;
		struct pxa_rttask down;
	} chans[MAX_CHANS+1];
	struct pxa_rttask frame;
};

static struct {
	int npins;
	char persistent;
	/* Request to stop the real-time tasks.
	 * Written once from kernel context, then decremented by RT tasks
	 * while the kernel watches.
	 * This assumes that reads and writes are atomic on ARM.  */
	unsigned char stopping;
	struct pin pins[MAX_PINS]; /* Must be last */
} lock;

/* Module API */

MODULE_DESCRIPTION("PWM/PPM signal generator for PXA255 GPIO");
MODULE_AUTHOR("pascal@pabr.org");
MODULE_LICENSE("GPL");

static int npins = 1;
module_param(npins, int, 0400);
MODULE_PARM_DESC(npins, "Number of signals (excl servo outputs)");

static int mode[MAX_PINS] = { [0 ... MAX_PINS-1] = PPM };
module_param_array(mode, int, NULL, 0400);
MODULE_PARM_DESC(mode, "0,1:PPM 2:DCM");

static int gpio[MAX_PINS] = { [0 ... MAX_PINS-1] = -1 };
module_param_array(gpio, int, NULL, 0400);
MODULE_PARM_DESC(gpio, "Multi-channel GPIO pins");

static int nchans[MAX_PINS] = { [0 ... MAX_PINS-1] = 1 };
module_param_array(nchans, int, NULL, 0400);
MODULE_PARM_DESC(nchans, "Number of channels");

static int tmin[MAX_PINS] = { [0 ... MAX_PINS-1] = 1000 };
module_param_array(tmin, int, NULL, 0400);
MODULE_PARM_DESC(tmin, "Pulse width for 0% (microseconds)");

static int tmax[MAX_PINS] = { [0 ... MAX_PINS-1] = 2000 };
module_param_array(tmax, int, NULL, 0400);
MODULE_PARM_DESC(tmax, "Pulse width for 100% (microseconds)");

static int tpause[MAX_PINS] = { [0 ... MAX_PINS-1] = 500 };
module_param_array(tpause, int, NULL, 0400);
MODULE_PARM_DESC(tpause, "Time between pulses (microseconds)");

static int tsync[MAX_PINS] = { [0 ... MAX_PINS-1] = 12000 };
module_param_array(tsync, int, NULL, 0400);
MODULE_PARM_DESC(tsync, "Time between pulse sets (microseconds)");

static int servo[MAX_PINS*MAX_CHANS]
= { [0 ... MAX_PINS*MAX_CHANS-1] = -1 };
module_param_array(servo, int, NULL, 0400);
MODULE_PARM_DESC(servo, "Time between pulse sets (microseconds)");

static int persistent = 0;
module_param(persistent, int, 0400);
MODULE_PARM_DESC(persistent, "Generate signal while device is closed");

/**********************************************************************/
/*                         File operations                            */
/**********************************************************************/

static ssize_t opwm_pin_write(struct file *file, const char *buf,
			      size_t count, loff_t *offp) {    
	struct pin *pin = file->private_data;
	int size = pin->nchans * sizeof(val_t);
	struct chan *chan = pin->chans;
	val_t *buf1 = (val_t*)buf;
	if ( count < size ) return -EINVAL;
	for ( count=pin->nchans; count--; ++chan,++buf1 )
		if ( get_user(chan->val, buf1) ) return -EFAULT;
	pin->status |= VALID;
	return size;
}

static ssize_t opwm_chan_write(struct file *file, const char *buf,
			       size_t count, loff_t *offp) {    
	struct chan *chan = file->private_data;
	if ( count < sizeof(val_t) ) return -EINVAL;
	if ( get_user(chan->val, (val_t*)buf) ) return -EFAULT;
	return sizeof(val_t);
}

static int opwm_pin_release(struct inode *inode, struct file *file) {
	struct pin *pin = file->private_data;
	pin->status &= ~IN_USE;
	if ( ! lock.persistent ) pin->status &= ~VALID;
	return 0;
}

static int opwm_chan_release(struct inode *inode, struct file *file) {
	struct chan *chan = file->private_data;
	chan->in_use = 0;
	return 0;
}

static struct file_operations opwm_pin_fops = {
	.owner   = THIS_MODULE,
	.write   = opwm_pin_write,
	.release = opwm_pin_release,
};

static struct file_operations opwm_chan_fops = {
	.owner   = THIS_MODULE,
	.write   = opwm_chan_write,
	.release = opwm_chan_release,
};

static int opwm_open(struct inode *inode, struct file *file) {
	int pin_num = PIN_NUM(inode->i_rdev);
	int chan_num = CHAN_NUM(inode->i_rdev);
	struct pin *pin;
	if ( pin_num<0 || pin_num>=lock.npins )
		return -ENODEV;
	pin = &lock.pins[pin_num];
	if ( chan_num == 15 ) {
		if ( pin->status & IN_USE ) return -EBUSY;
		pin->status |= IN_USE;
		file->f_op = &opwm_pin_fops;
		file->private_data = pin;
	} else {
		struct chan *chan;
		if ( chan_num<0 || chan_num>=pin->nchans )
			return -ENODEV;
		chan = &pin->chans[chan_num];
		if ( chan->in_use ) return -EBUSY;
		chan->in_use = 1;
		file->f_op = &opwm_chan_fops;
		file->private_data = chan;
	}
	return 0;
}

static struct file_operations opwm_fops = {
	.owner = THIS_MODULE,
	.open  = opwm_open,
};


/**********************************************************************/
/*                                PWM                                 */
/**********************************************************************/

#define gpio_set(bit)   GPSR(bit) = GPIO_bit(bit)
#define gpio_clear(bit) GPCR(bit) = GPIO_bit(bit)

void ILOCK rt_pin_start_frame(pxa_rtc_t t, void *data) {
	struct pin *pin = (struct pin *)data;
	int c;
	if ( lock.stopping ) {
		--lock.stopping;
		return;
	}
	if ( ! (pin->status&VALID) ) {
		pin->frame.base = RTSCHED_REL_NOW;
		pin->frame.next = NULL;
		rt_schedule_taskset(t, &pin->frame);
		return;
	}

	switch ( pin->mode ) {
	case PPM: {
		struct chan *chan;
		pxa_rtc_t dt, pulse;
		for ( c=0; c<pin->nchans; ++c ) {
			chan = &pin->chans[c];
			dt = pin->tmax - pin->tmin;
			pulse = pin->tmin + ((dt*chan->val) >> 16);
			chan->up.next = &chan->down;
			chan->down.next = &pin->chans[c+1].up;
			pin->chans[c+1].up.tmin = pulse;
			pin->chans[c+1].up.tmax = pulse;
		}
		
		chan = &pin->chans[pin->nchans];
		chan->up.next = &chan->down;
		chan->down.next = &pin->frame;
		pin->frame.base = RTSCHED_REL_PRED;
		pin->frame.next = NULL;
		rt_schedule_taskset(t, &pin->chans[0].up);
		break;
	}
	case PWM:
		break;
	case DCM:
		break;
	}
}

void ILOCK rt_chan_up(pxa_rtc_t t, void *data) {
	struct chan *chan = (struct chan *)data;
	struct pin *pin = chan->pin;
	/* End previous servo pulse, if any */
	if ( chan != &pin->chans[0] && (chan-1)->gpio >= 0 )
		gpio_clear((chan-1)->gpio);
	/* Start pulses. */
	if ( pin->gpio >= 0 ) gpio_set(pin->gpio);
	if ( chan->gpio >= 0 ) gpio_set(chan->gpio);
}

void ILOCK rt_chan_down(pxa_rtc_t t, void *data) {
	struct chan *chan = (struct chan *)data;
	struct pin *pin = chan->pin;
	/* End multiplexed pulse. */
	if ( pin->gpio >= 0 ) gpio_clear(pin->gpio);
}

#if USE_CACHE_LOCKING
void ILOCK i_lock_end(void) { }
#endif

/**********************************************************************/
/*                             Module                                 */
/**********************************************************************/

static struct class *class;

static int __init pxa_opwm_init(void) {
	int ret;
	int p, sp;

	memset(&lock, 0, sizeof(lock));

	/* Check config. */

	lock.npins = npins;
	if ( lock.npins<1 || lock.npins>MAX_PINS ) return -EINVAL;

	lock.persistent = persistent;
	lock.stopping = 0;

	for ( sp=0,p=0; p<lock.npins; ++p ) {
		struct pin *pin = &lock.pins[p];
		int c;
		pxa_rtc_t tp, ts;
		pin->mode = mode[p];
		pin->gpio = gpio[p];
		pin->status = 0;
		pin->nchans = nchans[p];
		switch ( mode[p] ) {
		case PPM:
			tmin[p] -= tpause[p];
			tmax[p] -= tpause[p];
			tsync[p] -= tpause[p];
			break;
		case PWM:
			printk(NAME ": For PWM-only, use mode=1, gpio=-1"
			       " and servo=<GPIO>.\n");
			return -EINVAL;
		default:
			printk(NAME ": mode %d not supported.", mode[p]);
			break;
		}
		pin->tmin = us2clocks(tmin[p]);
		pin->tmax = us2clocks(tmax[p]);
		tp = us2clocks(tpause[p]);
		ts = us2clocks(tsync[p]);
		printk(NAME "%d on gpio %d mode=%d "
		       "nchans=%d tmin=%d tmax=%d tpause=%d tsync=%d\n",
		       p, pin->gpio, pin->mode, pin->nchans,
		       pin->tmin, pin->tmax, tp, ts);
		if ( pin->gpio<-1 || pin->gpio>MAX_GPIO ||
		     pin->nchans<1 || pin->nchans>MAX_CHANS ||
		     (s32)pin->tmin<=0 || (s32)pin->tmax<=(s32)pin->tmin ||
		     (s32)tp<=0 || (s32)ts<=UPDOWN_DURATION )
			return -EINVAL;
		u32 tprep;
		if ( ts > STARTFRAME_DURATION+UPDOWN_DURATION ) {
			// If we have plenty of time, wait until as
			// late as possible to have recent data.
			tprep = STARTFRAME_DURATION;
		} else {
			// Otherwise, wait the bare minimum to get
			// consistent timings.
			tprep = ts - UPDOWN_DURATION;
			printk(NAME ": Warning: tsync may be too short\n");
		}
		/* nchans + one dummy channel for sync */
		for ( c=0; c<pin->nchans+1; ++c ) {
			int cgpio = (c<pin->nchans) ? servo[sp++] : -1;
			struct chan *chan = &pin->chans[c];
			chan->pin = pin;
			if ( cgpio<-1 || cgpio>MAX_GPIO )
				return -EINVAL;
			chan->gpio = cgpio;
			if ( cgpio >= 0 )
				printk(NAME"%d-%d on gpio %d\n",
				       p, c, cgpio);
			chan->up.rt_code = rt_chan_up;
			chan->up.data = chan;
			chan->up.duration = UPDOWN_DURATION;
			if ( c == 0 ) {
				chan->up.base = RTSCHED_REL_NOW;
				chan->up.tmin = tprep;
				chan->up.tmax = RTSCHED_ANYTIME;
			} else { 
				chan->up.base = RTSCHED_REL_PRED;
				/* Timings are updated in start_frame(). */
			}
			chan->down.rt_code = rt_chan_down;
			chan->down.data = chan;
			chan->down.duration = UPDOWN_DURATION;
			chan->down.base = RTSCHED_REL_PRED;
			chan->down.tmin = tp;
			chan->down.tmax = tp;
		}
		pin->frame.rt_code = rt_pin_start_frame;
		pin->frame.data = pin;
		pin->frame.duration = STARTFRAME_DURATION;
		pin->frame.base = RTSCHED_REL_NOW; /* First time only */
		/* Wakeup some time before the first pulse. */
		pin->frame.tmin = ts - tprep;
		pin->frame.tmax = RTSCHED_ANYTIME;
	}

	ret = register_chrdev(dev_major, NAME, &opwm_fops);
	if ( ret < 0 ) goto fail1;
	if ( ! dev_major ) dev_major = ret;

	class = class_create(THIS_MODULE, NAME);
	if ( ! class ) { ret = -EBUSY; goto fail2; }
	for ( p=0; p<lock.npins; ++p ) {
		struct pin *pin = &lock.pins[p];
		class_device_create(class, NULL, MKDEV(dev_major,p*16+15),
				    NULL, NAME"%d", p);
		int c;
		for ( c=0; c<pin->nchans; ++c ) {
			class_device_create(class, NULL,
					    MKDEV(dev_major,p*16+c),
					    NULL, NAME"%d-%d", p, c);
		}
	}

	/* Setup gpio */

	for ( p=0; p<lock.npins; ++p ) {
		struct pin *pin = &lock.pins[p];
		int c, gpio = pin->gpio;
		if ( gpio >= 0 ) {
			pxa_gpio_mode(gpio | GPIO_OUT);
			gpio_clear(gpio);           /* Output 0 */
		}
		for ( c=0; c<pin->nchans; ++c ) {
			int cgpio = pin->chans[c].gpio;
			if ( cgpio >= 0 ) {
				pxa_gpio_mode(cgpio | GPIO_OUT);
				gpio_clear(cgpio); /* Output 0 */
			}
		}
			
	}

	/* Lock cache. */
#if USE_CACHE_LOCKING
	printk(NAME ": code %p-%p, data %p - %p\n",
	       i_lock_begin, i_lock_end,
	       &lock, &lock.pins[lock.npins]);
	lock_i_cache(i_lock_begin, i_lock_end);
	lock_d_cache(&lock, &lock.pins[lock.npins]);
	/* Must lock constants that GCC puts in .text. */
	lock_d_cache(i_lock_begin, i_lock_end); /* TBD Overkill */
#endif
#if USE_CACHE_LOCKING && USE_TLB_LOCKING
	lock_i_tlb(i_lock_begin, i_lock_end);
	lock_d_tlb(&lock, &lock.pins[lock.npins]);
	lock_d_tlb((void*)&GPCR0, (void*)&GPCR0+1);  // GPIO regs
        /* Must lock constants that GCC puts in .text. */
	lock_d_tlb(i_lock_begin, i_lock_end);
#endif
	/* Spawn tasks, one per pin. */
	for ( p=0; p<lock.npins; ++p )
		lock.pins[p].frame.next =
			(p+1==lock.npins) ? NULL : &lock.pins[p+1].frame;
	ret = rtsched_spawn_tasks(&lock.pins[0].frame);
	if ( ret ) {
		printk(KERN_WARNING NAME ": Unable to spawn tasks.\n");
		goto fail3;
	}

	printk(NAME": Initialized, %d pin(s).\n", lock.npins);

	return 0;

fail3:
#if USE_CACHE_LOCKING && USE_TLB_LOCKING
	unlock_i_tlb();
	unlock_d_tlb();
#endif
#if USE_CACHE_LOCKING
	unlock_i_cache(); /* TBD: finer-grained. */
	unlock_d_cache();
#endif
	class_destroy(class);
fail2:  unregister_chrdev(dev_major, NAME);
fail1:  printk(KERN_WARNING NAME ": Aborting (%d).\n", ret);
	return ret;
}

static void __exit pxa_opwm_exit(void) {
	int p;
  	printk(KERN_WARNING NAME ": Exiting.\n");

	class_destroy(class);
	unregister_chrdev(dev_major, NAME);
	
	/* Disable GPIO output */
	for ( p=0; p<lock.npins; ++p ) {
		struct pin *pin = &lock.pins[p];
		int c, gpio = pin->gpio;
		if ( gpio >= 0 ) pxa_gpio_mode(gpio | GPIO_IN);
		for ( c=0; c<pin->nchans; ++c ) {
			int cgpio = pin->chans[c].gpio;
			if ( cgpio >= 0 ) pxa_gpio_mode(cgpio | GPIO_IN);
		}
	}

#if SAFE_RMMOD
	/* Shutdown RT before we mess with cache locks. */
	rtsched_stop();
#else
	/* Clean shutdown - Ask the RT tasks to terminate. */
	lock.stopping = lock.npins;
	while ( lock.stopping ) mdelay(1);
#endif

#if PROFILE
	for ( p=0; p<lock.npins; ++p ) {
		struct pin *pin = &lock.pins[p];
		int c;
		printk("pin%d dur_max=%d over=%d\n",
		       p, pin->frame.dur_max, pin->frame.overruns);
		for ( c=0; c<pin->nchans+1; ++c )
			printk(" %d-%d dur_max=%d,%d over=%d,%d\n",
			       p, c,
			       pin->chans[c].up.dur_max,
			       pin->chans[c].down.dur_max,
			       pin->chans[c].up.overruns,
			       pin->chans[c].down.overruns);
	}
#endif

#if USE_CACHE_LOCKING && USE_TLB_LOCKING
	unlock_i_tlb();
	unlock_d_tlb();
#endif
#if USE_CACHE_LOCKING
	unlock_i_cache(); /* TBD: finer-grained. */
	unlock_d_cache();
#endif
}

module_init(pxa_opwm_init);
module_exit(pxa_opwm_exit);
