/* -*- linux-c -*-
 *
 * Simple real-time scheduler for the PXA255
 * Copyright (c) 2005, 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/delay.h>

#include <asm/types.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/leds.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>
#include <asm/mach/time.h>
#include <asm/arch/pxa-regs.h>
#include <asm/cacheflush.h>

#include "pxa_rtsched.h"

/* Use IRQ instead of FIQ ?  (useful for debugging only) */
#define USE_IRQ 0

#define DEV_NAME "pxa_rtsched"

/* We use OS Match Register #1 of the PXA255. */
/* Change these if you want to use use OSMR2 instead. */
#define IRQ_RTSCHED  IRQ_OST1
#define OIER_RTSCHED OIER_E1
#define OSMR_RTSCHED OSMR1
#define OSSR_RTSCHED OSSR_M1

#define IRQ_EDGE  IRQ_GPIO_2_x

#ifdef CONFIG_SMP
#  error "pxa_rtsched does not support CONFIG_SMP"
#endif
#ifdef CONFIG_PREEMP
#  error "pxa_rtsched does not support CONFIG_PREEMPT"
#endif

#if USE_CACHE_LOCKING
#include "pxa_cache_lock.h"
#define ILOCK __attribute__((section(".lockable.text")))
void ILOCK i_lock_begin(void) { }
#else
#define ILOCK
#endif

/* ACK_LATENCY determines how fast we are able to
 * escape from race conditions after programming OSMR. */
#if USE_CACHE_LOCKING && USE_TLB_LOCKING
#  define ACK_LATENCY 3   // 1 clock per peripheral bus access
#else
#  define ACK_LATENCY 20  // Same + cache misses.
#endif

#define HW_IRQ(x) ((x)+PXA_IRQ_SKIP)

/**********************************************************************/
/* The spawner task                                                  */
/**********************************************************************/

#define SPAWNER_PERIOD (RTSCHED_FREQ>>7) /* 8 ms */
#define SPAWNER_DURATION 200

void rt_spawner(pxa_rtc_t t, void *data);

#define STACK_SIZE 4096
static int stack_locked = 1024; // Enough for 6-channel PPM
module_param(stack_locked, int, 0400);
MODULE_PARM_DESC(stack_locked, "# of bytes to lock into d-cache");

static struct {
        unsigned char fiq_stack[STACK_SIZE]; /* Must be first */
	struct pxa_rttask *next_task; /* Linked list of scheduled tasks */
	edge_handler_t eh_handler;
	struct pxa_rttask spawner_task;
	struct pxa_rttask *to_launch;
	int launch_result;
#if PROFILE_CACHE
	u32 pmn0;            /* PMN0 events during FIQ */
	u32 pmn1;            /* PMN1 events during FIQ */
	u32 pmn0_max;
	u32 pmn1_max;
#endif
#if PROFILE
	int n_intrs;
	int n_overruns;      /* # of times a task took too long */
	int n_overloads;     /* Multiple tasks in a single interrupt */
	int n_rejects;       /* # of scheduling failures */
	u32 time_in_intr;    /* Total time in interrupt */
	s32 tintr_max;       /* Max time in interrupt */
	s32 lat_total;       /* Total interrupt latency */
	s32 lat_min;         /* Min interrupt latency */
	s32 lat_max;         /* Max interrupt latency */
	s32 tlat_min;        /* Min task latency */
	s32 tlat_max;        /* Max task latency */
	s32 tdur_max;        /* Max task duration */
#define N_SLOTS 10
	int lat_slot[N_SLOTS];
#endif
} lock = {
	.spawner_task = {
		.rt_code  = rt_spawner,
		.data     = NULL,
		.duration = SPAWNER_DURATION,
		.base     = RTSCHED_REL_NOW,
		.tmin     = SPAWNER_PERIOD,
		.tmax     = RTSCHED_ANYTIME,
		.next     = NULL
	},
	.next_task = &lock.spawner_task
#if PROFILE
	,
	.tintr_max = -RTSCHED_FREQ,  /* Max time in interrupt */
	.lat_min   =  RTSCHED_FREQ,  /* Min interrupt latency */
	.lat_max   = -RTSCHED_FREQ,  /* Max interrupt latency */
	.tlat_min  =  RTSCHED_FREQ,  /* Min task latency */
	.tlat_max  = -RTSCHED_FREQ   /* Max task latency */
#endif
};

static u32 eh_fe[3];  /* Falling edges handled by us */
static u32 eh_re[3];  /* Rising edges handled by us */

void ILOCK rt_spawner(pxa_rtc_t t, void *data) {
	if ( lock.to_launch ) {
		lock.launch_result = rt_schedule_taskset(t, lock.to_launch);
		lock.to_launch = NULL;
	}
}

int rtsched_spawn_tasks(struct pxa_rttask *first) {
#if SAFE_RMMOD
	if ( ! lock.next_task ) {
		/* RT has been shut down. */
		printk(DEV_NAME ": Scheduler stopped, please rmmod now.\n");
		return -ESHUTDOWN;
	}
#endif
	lock.to_launch = first;
	/* Wait for the the spawner task to process our request. */
	while ( lock.to_launch ) msleep_interruptible(10);
	return lock.launch_result;
}

/**********************************************************************/
/*                      Hook for edge detection                       */
/**********************************************************************/

int rtsched_set_edge_handler(edge_handler_t h, u32 re[3], u32 fe[3]) {
	if ( lock.eh_handler ) return -EBUSY;
#if SAFE_RMMOD
	if ( ! lock.next_task ) {
		/* RT has been shut down. */
		printk(DEV_NAME ": Scheduler stopped, please rmmod now.\n");
		return -ESHUTDOWN;
	}
#endif

	lock.eh_handler = h;
	memcpy(eh_re, re, sizeof(eh_re));
	memcpy(eh_fe, fe, sizeof(eh_fe));
	for ( int i=0; i<3; ++i ) /* Clear any old interrupt */
		(&GEDR0)[i] = re[i] | fe[i]; 
	for ( int i=0; i<3; ++i ) { /* Enable */
		(&GRER0)[i] |= re[i];
		(&GFER0)[i] |= fe[i];
	}
	ICLR |= 1 << HW_IRQ(IRQ_EDGE);  /* Set level=FIQ */
	ICMR |= 1 << HW_IRQ(IRQ_EDGE);  /* Unmask */
	return 0;
}

void rtsched_remove_edge_handler(edge_handler_t h) {
	if ( h != lock.eh_handler ) {
		printk(DEV_NAME ": Tried to remove handler !\n");
		return;
	}
	for ( int i=0; i<3; ++i ) { /* Disable */
		(&GRER0)[i] &= ~eh_re[i];
		(&GFER0)[i] &= ~eh_fe[i];
	}
	for ( int i=0; i<3; ++i ) /* Clear */
		(&GEDR0)[i] = eh_re[i] | eh_fe[i];
	// ICMR &= ~(1 << HW_IRQ(IRQ_EDGE));  /* Safe not to mask */
	ICLR &= ~(1 << HW_IRQ(IRQ_EDGE));  /* Set level=IRQ */
	lock.eh_handler = NULL;
}

/**********************************************************************/
/* The scheduler                                                      */
/**********************************************************************/

static int ILOCK rt_sched(pxa_rtc_t min, pxa_rtc_t now, pxa_rtc_t pred,
			  struct pxa_rttask **pcur, struct pxa_rttask *new) {
	u32 tmin, tmax;
	int res;
	struct pxa_rttask *cur;
	if ( ! new ) return 0;
	switch ( new->base ) {
	case RTSCHED_ABS:  tmin=new->tmin; tmax=new->tmax; break;
	case RTSCHED_REL_NOW: tmin=now+new->tmin; tmax=now+new->tmax; break;
	case RTSCHED_REL_PRED: tmin=pred+new->tmin; tmax=pred+new->tmax; break;
	default: return -1;
	}
	/* Is it late ? */
	if ( (s32)(tmin-min) < 0 ) tmin = min;
	/* From now on we can ignore min. */
	/* Is it too late ? */
	if ( (s32)(tmax-tmin) < 0 ) return -1;
	cur = *pcur;
	/* Can we schedule at tmin, before the next task ? */
	min = tmin + new->duration;
	if ( !cur || (s32)(min-cur->tsched) < 0 ) {
		/* Insert. */
		struct pxa_rttask *new_next = new->next;
		*pcur = new;
		new->tsched = tmin;
		new->next = cur;
		/* Schedule remaining tasks. */
		res = rt_sched(min, now, new->tsched, &new->next, new_next);
		if ( ! res ) return 0;
		/* Undo. */
		*pcur = cur;
		new->next = new_next;
	}
	/* Last chance: schedule after. */
	min = cur->tsched + cur->duration;
	res = rt_sched(min, now, pred, &cur->next, new);
	return res;
}

int ILOCK rt_schedule_taskset(pxa_rtc_t now, struct pxa_rttask *first) {
	int res = rt_sched(now+10, now, now, &lock.next_task, first);
#if PROFILE
	if ( res ) ++lock.n_rejects;
#endif
	return res;
}

/**********************************************************************/

inline void rt_handler_timer(pxa_rtc_t t) {
	struct pxa_rttask *task = lock.next_task;
#if PROFILE
	s32 lat = t - task->tsched;
#endif

again:
	lock.next_task = task->next;
#if PROFILE
	//t = OSCR;
	s32 tlat = t - task->tsched;
#endif

	(*task->rt_code)(task->tsched, task->data);

#if PROFILE
	s32 tdur = OSCR - t;
	if ( tdur >= task->duration ) {
		++lock.n_overruns;
		++task->overruns;
	}
	if ( tdur > task->dur_max ) task->dur_max = tdur;
	if ( tdur > lock.tdur_max ) lock.tdur_max = tdur;
	if ( tlat < lock.tlat_min ) lock.tlat_min = tlat;
	if ( tlat > lock.tlat_max ) lock.tlat_max = tlat;
#endif

	rt_spawner(t, NULL);
	/* Make sure we have a next task. */
	if ( ! lock.next_task ) {
		lock.spawner_task.next = NULL;
		rt_schedule_taskset(task->tsched, &lock.spawner_task);
	}

	task = lock.next_task;
	/* Do we have time to ack without a race condition ? */
	OSMR_RTSCHED = task->tsched;
	if ( (s32)(task->tsched-OSCR) < ACK_LATENCY ) {
#if PROFILE
		++lock.n_overloads;
#endif
		while ( (s32)(t-task->tsched) < 0 ) t = OSCR;
		goto again;
	}
        /* Clear interrupt, especially those from overloads. */
	OSSR = OSSR_RTSCHED;

#if PROFILE
	lock.lat_total += lat;
	if ( lat < lock.lat_min ) lock.lat_min = lat;
	if ( lat > lock.lat_max ) lock.lat_max = lat;
	if ( lat>=0 && lat<N_SLOTS ) ++lock.lat_slot[lat];
	++lock.n_intrs;
#endif
}

void ILOCK rt_handler(pxa_rtc_t t) {
	//GPSR(60) = GPIO_bit(60);
#if PROFILE_CACHE
	u32 pmn0, pmn1; 
	__asm__ volatile("mrc p14, 0, %0, c2, c0, 0" : "=r"(pmn0));
	__asm__ volatile("mrc p14, 0, %0, c3, c0, 0" : "=r"(pmn1));
#endif
	u32 icpr = ICPR;
	/* We give priority to edge interrupts because the
	 * handler is likely to be smaller. */
	if ( icpr & (1<<HW_IRQ(IRQ_EDGE)) ) {
		/* Interrupt caused by GPIO edge. */
		(*lock.eh_handler)(t);
	} else if ( icpr & (1<<HW_IRQ(IRQ_RTSCHED)) ) {
		/* Interrupt caused by timer match. */
		rt_handler_timer(t);
	}
#if PROFILE
	int dt = OSCR - t;
	lock.time_in_intr += dt;
	if ( dt > lock.tintr_max ) lock.tintr_max = dt;
#endif
#if PROFILE_CACHE
	u32 pmn0bis, pmn1bis;
	__asm__ volatile("mrc p14, 0, %0, c2, c0, 0" : "=r"(pmn0bis));
	__asm__ volatile("mrc p14, 0, %0, c3, c0, 0" : "=r"(pmn1bis));
	pmn0 = pmn0bis - pmn0;
	pmn1 = pmn1bis - pmn1;
	if ( pmn0 > lock.pmn0_max ) lock.pmn0_max = pmn0;
	if ( pmn1 > lock.pmn1_max ) lock.pmn1_max = pmn1;
	lock.pmn0 += pmn0;
	lock.pmn1 += pmn1;
#endif
	//GPCR(60) = GPIO_bit(60);
}

#if USE_IRQ
irqreturn_t irq_handler(int irq, void *data, struct pt_regs *regs) {
	rt_handler(OSCR);
	return IRQ_HANDLED;
}
#else
/* For some reason CONFIG_FIQ is not enabled by CONFIG_GUMSTIX */
/* #include <asm/fiq.h> */
#endif

/* This code will copied to the FIQ vector.
 * Therefore we can't use relative adressing.
 */

void rt_fiq_stub(void), rt_fiq_stub_end(void);

void rt_fiq_stub_wrapper(void) {
	__asm__ volatile (
		"rt_fiq_stub:                                      \n\
                 @ FIQ registers: r8=OSCR, r9=handler, sp          \n\
                 @ Save OSCR ASAP for profiling.                   \n\
                 ldr r10, [r8]                                     \n\
                 @ Save registers.                                 \n\
                 stmdb sp!, {r0-r3,lr}                             \n\
                 mov r0, r10                                       \n\
                 @ Call the C function.                            \n\
                 mov lr, pc                                        \n\
                 mov pc, r9                                        \n\
                 @ Restore registers and return.                   \n\
                 ldmia sp!, {r0-r3,lr}                             \n\
                 subs pc, lr, #4                                   \n\
                 rt_fiq_stub_end:                                  \n\
                 ");
}

#if USE_CACHE_LOCKING
void ILOCK i_lock_end(void) { }
#endif

/**********************************************************************/
/*                             Module                                 */
/**********************************************************************/

#if ! USE_IRQ
static u32 *const fiq_vector = (u32*)0xffff001c;
static u32 fiq_saved_vector;
#endif

/* We should be using set_fiq_regs() from arch/arm/kernel/fiq.c
 * but CONFIG_ARCH_GUMSTIX does not currently enable CONFIG_FIQ.
 */

static void set_fiq_regs(u32 sp, u32 r8, u32 r9)
{
	u32 old_mode;
	__asm__ volatile (
		"mrs %[old_mode], cpsr      \n\
                msr cpsr_c, %[fiq_mode]     \n\
                mov r0, r0                  \n\
                mov r8, %[r8]               \n\
                mov r9, %[r9]               \n\
                mov sp, %[sp]               \n\
	        msr cpsr_c, %[old_mode]     \n\
	        mov r0, r0"
		: [old_mode]"=&r" (old_mode)
		: [r8]"r"(r8),
		  [r9]"r"(r9),
		  [sp]"r"(sp),
		  [fiq_mode]"I" (PSR_I_BIT|PSR_F_BIT|FIQ_MODE)
		: "r8","r9","r10","r11","r12","r13","r14","r15");
}

static int __init pxa_rtsched_init(void) {
	int ret;

#if ! USE_IRQ
	/* Install the FIQ stub code. */
	if ( stack_locked<0 || stack_locked>STACK_SIZE ) {
		printk(KERN_WARNING DEV_NAME ": Invalid stack size\n");
		ret = -EINVAL;
		goto error1;
	}
	set_fiq_regs((u32)&lock.fiq_stack[STACK_SIZE],
		     0xf2a00010, (u32)rt_handler);
	fiq_saved_vector = fiq_vector[0];
	int fiq_len = rt_fiq_stub_end - rt_fiq_stub;
	memcpy(fiq_vector, rt_fiq_stub, fiq_len);
	flush_icache_range((long)fiq_vector, (long)fiq_vector+fiq_len);
#if USE_CACHE_LOCKING
	unlock_i_cache();
	unlock_d_cache();
	printk(DEV_NAME ": code %p..%p, data %p..%p\n",
	       i_lock_begin, i_lock_end,
	       &lock.fiq_stack[STACK_SIZE-stack_locked], &lock+1);
	lock_i_cache(fiq_vector, (void*)fiq_vector+fiq_len);
	lock_i_cache(i_lock_begin, i_lock_end);
	lock_d_cache(&lock.fiq_stack[STACK_SIZE-stack_locked], &lock+1);
	/* Must lock constants that GCC puts in .text. */
	lock_d_cache(i_lock_begin, i_lock_end); /* TBD Overkill */
#endif
#if USE_CACHE_LOCKING && USE_TLB_LOCKING
	unlock_i_tlb();
	unlock_d_tlb();
	lock_i_tlb(fiq_vector, (void*)fiq_vector+fiq_len);
	lock_i_tlb(i_lock_begin, i_lock_end);
	lock_d_tlb(&lock.fiq_stack[STACK_SIZE-stack_locked], &lock+1);
	lock_d_tlb((void*)&OSSR, (void*)&OSSR+1);  // Timer regs
	lock_d_tlb((void*)&ICPR, (void*)&ICPR+1);   // Int cont regs
        /* Must lock constants that GCC puts in .text. */
	lock_d_tlb(i_lock_begin, i_lock_end);
#endif
#endif /* USE_IRQ */

#if PROFILE_CACHE
	// PMN0 = TLB misses, PMN1 = Cache misses
	//u32 pmnc_on = 1 | (0x03<<12) | (0x00<<20); // I
	u32 pmnc_on = 1 | (0x04<<12) | (0x0b<<20); // D
	__asm__ volatile("mcr p14, 0, %0, c0, c0, 0" : : "r"(pmnc_on));
#endif

	/* Start the spawner */
	lock.spawner_task.tsched = OSCR + SPAWNER_PERIOD;
	OSMR_RTSCHED = lock.spawner_task.tsched;
      	OSSR = OSSR_RTSCHED;               /* Clear interrupt */
	OIER |= OIER_RTSCHED;              /* Enable interrupt */
#if USE_IRQ
	ret = request_irq(IRQ_RTSCHED, &irq_handler,
			  SA_INTERRUPT, DEV_NAME, THIS_MODULE);
	if ( ret ) {
		printk(KERN_WARNING DEV_NAME ": Unable to allocate irq\n");
		goto error1;
	}
#else
	ICLR |= 1 << HW_IRQ(IRQ_RTSCHED);  /* Set level=FIQ */
	ICMR |= 1 << HW_IRQ(IRQ_RTSCHED);  /* Unmask */
#endif

	printk(DEV_NAME ": Initialized.\n");
	return 0;

 error1:
	printk(KERN_WARNING DEV_NAME ": Aborting.\n");
	return ret;
}

void rtsched_stop(void) {
#if PROFILE_CACHE
	/* Unlocking will mess the statistics, so stop PMN now. */
	u32 pmnc_off = 0;
	__asm__ volatile("mcr p14, 0, %0, c0, c0, 0" : : "r"(pmnc_off));
#endif
	OIER &= ~OIER_RTSCHED;             /* Disable interrupt */
      	OSSR = OSSR_RTSCHED;               /* Clear interrupt */
#if USE_IRQ
	free_irq(IRQ_RTSCHED, THIS_MODULE);
#else
	ICMR &= ~(1<<HW_IRQ(IRQ_RTSCHED)); /* Mask */
	ICLR &= ~(1<<HW_IRQ(IRQ_RTSCHED)); /* Set level=IRQ */
	fiq_vector[0] = fiq_saved_vector;  /* Restore vector */
#if USE_CACHE_LOCKING && USE_TLB_LOCKING
	unlock_i_tlb();
	unlock_d_tlb();
#endif
#if USE_CACHE_LOCKING
	unlock_i_cache();
	unlock_d_cache();
#endif
#endif /* USE_IRQ */

	lock.next_task = NULL;
	if ( lock.eh_handler ) rtsched_remove_edge_handler(lock.eh_handler);
}

static void __exit pxa_rtsched_exit(void) {
	printk(DEV_NAME ": Exiting.\n");

	rtsched_stop();

#if PROFILE_CACHE
	u32 pmnc_off = 0;
	__asm__ volatile("mcr p14, 0, %0, c0, c0, 0" : : "r"(pmnc_off));
	printk("PMN0=%u (max %u) PMN1=%u (max %u) ",
	       lock.pmn0, lock.pmn0_max,
	       lock.pmn1, lock.pmn1_max);
#endif

#if PROFILE
	if ( lock.n_intrs ) {
		printk("%d rejects, %d overruns, %d overloads\n",
		       lock.n_rejects, lock.n_overruns, lock.n_overloads);

		printk("%d interrupts, avg=%d, max=%d\n",
		       lock.n_intrs,
		       lock.time_in_intr/lock.n_intrs, lock.tintr_max);
		printk("Tasks: lat=%d..%d dur_max=%d\n",
		       lock.tlat_min, lock.tlat_max, lock.tdur_max);
		printk("Spawner: dur_max=%d over=%d\n",
		       lock.spawner_task.dur_max, lock.spawner_task.overruns);
		printk("Interrupt latency: %d..%d, avg=%d\n",
		       lock.lat_min, lock.lat_max,
		       lock.lat_total/lock.n_intrs);
		for ( int i=0; i<N_SLOTS; ++i ) {
			printk("%3d %10d ", i, lock.lat_slot[i]);
			for ( int j=lock.lat_slot[i]*60/lock.n_intrs; j--; )
				printk("#");
			printk("\n");
		}
		printk("\n");
	}
#endif
}

module_init(pxa_rtsched_init);
module_exit(pxa_rtsched_exit);

EXPORT_SYMBOL(rtsched_spawn_tasks);
EXPORT_SYMBOL(rtsched_stop);
EXPORT_SYMBOL(rt_schedule_taskset);

EXPORT_SYMBOL(rtsched_set_edge_handler);
EXPORT_SYMBOL(rtsched_remove_edge_handler);

MODULE_DESCRIPTION("Simple real-time scheduler for the PXA255");
MODULE_AUTHOR("pascal@pabr.org");
MODULE_LICENSE("GPL");
