/* -*- linux-c -*-
 *
 * Simple real-time scheduler for the PXA255
 * Copyright (c) 2005, 2006 pascal@pabr.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

/* This module meets real time constraints with bounded latency,
 * provided that:
 *
 * - Tasks terminate within the time they committed to
 *   (including overhead for rescheduling, which can be costly).
 *
 * - No one else uses or disables FIQs.
 *
 * - The system bus is not loaded by other hardware (DMA, ...).
 *   Even if all code and data is locked into the cache of the PXA255,
 *   we must go though the system bus to access the registers of
 *   the interrupt controller, timer, etc.
 *
 * A hook for enabling edge detection is provided.
 * Real-time tasks may be delayed if it is used.
 *
 */

#ifndef PXA_RTSCHED_H
#define PXA_RTSCHED_H

/* Generate statistics ? */
#define PROFILE 1
/* Detect cache/tlb misses ? */
#define PROFILE_CACHE 1

/* Use ARM cache locking to reduce latency ?
 * Note: Disabling this may trigger a potential problem related to
 * modules being vmalloc'ed and therefore not always mapped in VM.
 */
#define USE_CACHE_LOCKING 1
/* Also lock the TLB ? */
#define USE_TLB_LOCKING 1

/* Make rmmod safer ?  If 1, the whole module stack must be unloaded. */
#define SAFE_RMMOD 1

typedef u32 pxa_rtc_t;

enum pxa_rtbase { RTSCHED_ABS, RTSCHED_REL_NOW, RTSCHED_REL_PRED } ;

#define RTSCHED_FREQ 3686400

#define RTSCHED_ANYTIME 1000000000

struct pxa_rttask {
	pxa_rtc_t tsched;
	struct pxa_rttask *next;
	/* The remaining fields are never altered by the scheduler. */
	void (*rt_code)(pxa_rtc_t t, void *data);
	void *data;
	pxa_rtc_t duration;    /* Upper bound on execution time */
	pxa_rtc_t tmin;        /* Requested time of execution, min */
	pxa_rtc_t tmax;        /* Requested time of execution, max */
	char base;
#if PROFILE
	u8 overruns;           /* # of times the task took too long */
	u8 dur_max;            /* Max time spent in task */
#endif
};

/* Called from kernel context (user or IRQ).
 * Not thread-safe.
 * Slow (new tasks are polled by a realtime task).
 */
int rtsched_spawn_tasks(struct pxa_rttask *first);
void rtsched_stop(void);

/* Called from RT context */
int rt_schedule_taskset(pxa_rtc_t now, struct pxa_rttask *first);

/* Hook for GPIO edge detection. */
/* Note: Processing a edge interrupt will delay RT tasks. */
typedef void (*edge_handler_t)(pxa_rtc_t t);
int rtsched_set_edge_handler(edge_handler_t, u32 rising[3], u32 falling[3]);
void rtsched_remove_edge_handler(edge_handler_t);

inline u32 us2clocks(u64 us) {
	us = us*RTSCHED_FREQ + 500000;
	do_div(us, 1000000);
	return us;
}

#endif /* PXA_RTSCHED_H */
